import os
from typing import Optional, Tuple, Generator, TYPE_CHECKING, BinaryIO, Dict
import numpy
import copy

import amulet_nbt

from amulet.api.data_types import (
    AnyNDArray,
    VersionNumberAny,
    Dimension,
    PlatformType,
    PointCoordinates,
    ChunkCoordinates,
)
from amulet.api.registry import BlockManager
from amulet.api.wrapper import StructureFormatWrapper
from amulet.api.chunk import Chunk
from amulet.api.selection import SelectionBox, SelectionGroup
from amulet.api.errors import ObjectReadError, ObjectWriteError, ChunkDoesNotExist
from .interface import (
    JavaSchematicInterface,
    BedrockSchematicInterface,
    SchematicInterface,
)
from .data_types import BlockArrayType, BlockDataArrayType
from .chunk import SchematicChunk

if TYPE_CHECKING:
    from amulet.api.wrapper import Translator, Interface

java_interface = JavaSchematicInterface()
bedrock_interface = BedrockSchematicInterface()


class SchematicFormatWrapper(StructureFormatWrapper):
    def __init__(self, path: str):
        super().__init__(path)
        self._chunks: Dict[
            ChunkCoordinates,
            Tuple[SelectionBox, BlockArrayType, BlockDataArrayType, list, list],
        ] = {}

    def _create(self, overwrite: bool, **kwargs):
        if not overwrite and os.path.isfile(self.path):
            raise ObjectWriteError(f"There is already a file at {self.path}")
        if self._platform == "bedrock":
            self._version = (1, 12, 0)
        else:
            self._platform = "java"
            self._version = (1, 12, 2)
        self._chunks = {}

    def open_from(self, f: BinaryIO):
        schematic = amulet_nbt.load(f)
        if "BlockData" in schematic:
            raise ObjectReadError("This file is not a legacy schematic file.")
        materials = schematic.get("Materials", amulet_nbt.TAG_String()).value
        if materials == "Alpha":
            self._platform = "java"
            self._version = (1, 12, 2)
        elif materials == "Pocket":
            self._platform = "bedrock"
            self._version = (1, 12, 0)
        else:
            raise Exception(
                f'"{materials}" is not a supported platform for a schematic file.'
            )
        self._chunks = {}
        selection_box = SelectionBox(
            (0, 0, 0),
            (
                schematic["Width"].value,
                schematic["Height"].value,
                schematic["Length"].value,
            ),
        )
        self._selection = SelectionGroup(selection_box)
        entities: amulet_nbt.TAG_List = schematic.get("Entities", amulet_nbt.TAG_List())
        block_entities: amulet_nbt.TAG_List = schematic.get(
            "TileEntities", amulet_nbt.TAG_List()
        )
        blocks: numpy.ndarray = (
            schematic["Blocks"].value.astype(numpy.uint8).astype(numpy.uint16)
        )
        if "AddBlocks" in schematic:
            add_blocks = schematic["AddBlocks"]
            blocks = (
                blocks
                + (
                    numpy.concatenate([[(add_blocks & 0xF0) >> 4], [add_blocks & 0xF]])
                    .T.ravel()
                    .astype(numpy.uint16)
                    << 8
                )[: blocks.size]
            )
        max_point = selection_box.max
        temp_shape = (max_point[1], max_point[2], max_point[0])
        blocks = numpy.transpose(blocks.reshape(temp_shape), (2, 0, 1))  # YZX => XYZ
        data = numpy.transpose(
            schematic["Data"].value.reshape(temp_shape), (2, 0, 1)
        ).astype(numpy.uint8)
        for cx, cz in selection_box.chunk_locations():
            box = SelectionBox(
                (cx * self.sub_chunk_size, 0, cz * self.sub_chunk_size),
                (
                    min((cx + 1) * self.sub_chunk_size, selection_box.size_x),
                    selection_box.size_y,
                    min((cz + 1) * self.sub_chunk_size, selection_box.size_z),
                ),
            )
            self._chunks[(cx, cz)] = (box, blocks[box.slice], data[box.slice], [], [])
        for e in block_entities:
            if all(key in e for key in ("x", "y", "z")):
                x, y, z = e["x"].value, e["y"].value, e["z"].value
                if (x, y, z) in selection_box:
                    cx = x >> 4
                    cz = z >> 4
                    self._chunks[(cx, cz)][3].append(e)
        for e in entities:
            if "Pos" in e:
                pos: PointCoordinates = tuple(
                    map(lambda t: float(t.value), e["Pos"].value)
                )
                if pos in selection_box:
                    cx = int(pos[0]) >> 4
                    cz = int(pos[2]) >> 4
                    self._chunks[(cx, cz)][4].append(e)

    @staticmethod
    def is_valid(path: str) -> bool:
        """
        Returns whether this format is able to load the given object.

        :param path: The path of the object to load.
        :return: True if the world can be loaded by this format, False otherwise.
        """
        return os.path.isfile(path) and path.endswith(".schematic")

    @property
    def valid_formats(self) -> Dict[PlatformType, Tuple[bool, bool]]:
        return {"java": (True, True), "bedrock": (True, True)}

    @property
    def extensions(self) -> Tuple[str, ...]:
        return (".schematic",)

    def _get_interface(self, raw_chunk_data=None) -> SchematicInterface:
        if self._platform == "java":
            return java_interface
        elif self._platform == "bedrock":
            return bedrock_interface
        else:
            raise Exception(f"{self._platform} is not a supported platform")

    def _get_interface_and_translator(
        self, raw_chunk_data=None
    ) -> Tuple["Interface", "Translator", VersionNumberAny]:
        interface = self._get_interface(raw_chunk_data)
        translator, version_identifier = interface.get_translator(
            self.max_world_version, raw_chunk_data, self.translation_manager
        )
        return interface, translator, version_identifier

    def save_to(self, f: BinaryIO):
        if self._platform == "java":
            materials = "Alpha"
        elif self._platform == "bedrock":
            materials = "Pocket"
        else:
            raise ObjectWriteError(
                f'"{self._platform}" is not a supported platform for a schematic file.'
            )

        selection = self._selection.selection_boxes[0]

        data = amulet_nbt.NBTFile(
            amulet_nbt.TAG_Compound(
                {
                    "TileTicks": amulet_nbt.TAG_List(),
                    "Width": amulet_nbt.TAG_Short(selection.size_x),
                    "Height": amulet_nbt.TAG_Short(selection.size_y),
                    "Length": amulet_nbt.TAG_Short(selection.size_z),
                    "Materials": amulet_nbt.TAG_String(materials),
                }
            ),
            "Schematic",
        )

        entities = []
        block_entities = []
        blocks = numpy.zeros(
            selection.shape, dtype=numpy.uint16
        )  # only 12 bits are actually used at most
        block_data = numpy.zeros(
            selection.shape, dtype=numpy.uint8
        )  # only 4 bits are used

        for (
            selection_,
            blocks_,
            data_,
            block_entities_,
            entities_,
        ) in self._chunks.values():
            if selection_.intersects(selection):
                box = selection_.create_moved_box(selection.min, subtract=True)
                blocks[box.slice] = blocks_
                block_data[box.slice] = data_
                for be in block_entities_:
                    coord_type = be["x"].__class__
                    be["x"] = coord_type(be["x"] - selection.min_x)
                    be["y"] = coord_type(be["y"] - selection.min_y)
                    be["z"] = coord_type(be["z"] - selection.min_z)
                    block_entities.append(be)
                for e in entities_:
                    coord_type = e["Pos"][0].__class__
                    e["Pos"][0] = coord_type(e["Pos"][0] - selection.min_x)
                    e["Pos"][1] = coord_type(e["Pos"][1] - selection.min_y)
                    e["Pos"][2] = coord_type(e["Pos"][2] - selection.min_z)
                    entities.append(e)

        data["Entities"] = amulet_nbt.TAG_List(entities)
        data["TileEntities"] = amulet_nbt.TAG_List(block_entities)
        data["Data"] = amulet_nbt.TAG_Byte_Array(
            numpy.transpose(block_data, (1, 2, 0))  # XYZ => YZX
        )
        data["Blocks"] = amulet_nbt.TAG_Byte_Array(
            numpy.transpose((blocks & 0xFF).astype(numpy.uint8), (1, 2, 0))
        )
        if numpy.max(blocks) > 0xFF:
            add_blocks = (numpy.transpose(blocks & 0xF00, (1, 2, 0)) >> 8).ravel()
            data["AddBlocks"] = amulet_nbt.TAG_Byte_Array(
                add_blocks[::2] + (add_blocks[1::2] << 4)
            )
        data.save_to(f)

    def _close(self):
        """Close the disk database"""
        self._chunks.clear()

    def unload(self):
        """Unload data stored in the Format class"""
        pass

    def all_chunk_coords(
        self, dimension: Optional[Dimension] = None
    ) -> Generator[Tuple[int, int], None, None]:
        """A generator of all chunk coords"""
        yield from self._chunks.keys()

    def has_chunk(self, cx: int, cz: int, dimension: Dimension) -> bool:
        return (cx, cz) in self._chunks

    def _pack(
        self,
        chunk: "Chunk",
        translator: "Translator",
        chunk_version: VersionNumberAny,
    ) -> Tuple["Chunk", AnyNDArray]:
        version = self.translation_manager.get_version(
            *translator.translator_key(chunk_version)
        )
        palette = []
        for entry in chunk.block_palette.blocks():
            b = version.block.block_to_ints(entry)
            if b is None:
                b = (0, 0)
            palette.append(b)
        return (
            chunk,
            numpy.array(palette),
        )

    def _encode(
        self,
        chunk: Chunk,
        chunk_palette: AnyNDArray,
        interface: SchematicInterface,
    ):
        return interface.encode(
            chunk,
            chunk_palette,
            self.max_world_version,
            SelectionBox.create_chunk_box(chunk.cx, chunk.cz).intersection(
                self._selection.to_box()
            ),
        )

    def _unpack(
        self,
        translator: "Translator",
        game_version: VersionNumberAny,
        chunk: "Chunk",
        chunk_palette: AnyNDArray,
    ) -> "Chunk":
        version = self.translation_manager.get_version(
            *translator.translator_key(game_version)
        )
        palette = chunk._block_palette = BlockManager()
        lut = numpy.array(
            [
                palette.get_add_block(version.block.ints_to_block(block, data))
                for block, data in chunk_palette
            ]
        )
        if len(palette.blocks()) != len(chunk_palette):
            # if a blockstate was defined twice
            for cy in chunk.blocks.sub_chunks:
                chunk.blocks.add_sub_chunk(cy, lut[chunk.blocks.get_sub_chunk(cy)])

        return chunk

    def _delete_chunk(self, cx: int, cz: int, dimension: Optional[Dimension] = None):
        if (cx, cz) in self._chunks:
            del self._chunks[(cx, cz)]

    def _put_raw_chunk_data(
        self,
        cx: int,
        cz: int,
        section: SchematicChunk,
        dimension: Optional[Dimension] = None,
    ):
        """
        Actually stores the data from the interface to disk.
        """
        self._chunks[(cx, cz)] = copy.deepcopy(
            (
                section.selection,
                section.blocks,
                section.data,
                section.block_entities,
                section.entities,
            )
        )

    def _get_raw_chunk_data(
        self, cx: int, cz: int, dimension: Optional[Dimension] = None
    ) -> SchematicChunk:
        """
        Return the raw data as loaded from disk.

        :param cx: The x coordinate of the chunk.
        :param cz: The z coordinate of the chunk.
        :param dimension: The dimension to load the data from.
        :return: The raw chunk data.
        """
        if (cx, cz) in self._chunks:
            return SchematicChunk(*copy.deepcopy(self._chunks[(cx, cz)]))
        else:
            raise ChunkDoesNotExist



import os
from typing import Optional, List, Tuple, Dict, Generator, TYPE_CHECKING, BinaryIO, Any
import numpy
from io import BytesIO
import struct
import copy

import amulet_nbt

from amulet import log
from amulet.api.data_types import AnyNDArray, VersionNumberAny, Dimension, PlatformType
from amulet.api.registry import BlockManager
from amulet.api.wrapper import StructureFormatWrapper
from amulet.api.chunk import Chunk
from amulet.api.selection import SelectionGroup, SelectionBox
from amulet.api.errors import ChunkDoesNotExist, ObjectWriteError

from .section import ConstructionSection
from .interface import Construction0Interface, ConstructionInterface
from .util import (
    unpack_palette,
    parse_entities,
    parse_block_entities,
    serialise_entities,
    serialise_block_entities,
    find_fitting_array_type,
    pack_palette,
)

if TYPE_CHECKING:
    from amulet.api.wrapper import Translator, Interface

construction_0_interface = Construction0Interface()

INT_STRUCT = struct.Struct(">I")
SECTION_ENTRY_TYPE = numpy.dtype(
    [
        ("sx", "i4"),
        ("sy", "i4"),
        ("sz", "i4"),
        ("shapex", "i1"),
        ("shapey", "i1"),
        ("shapez", "i1"),
        ("position", "i4"),
        ("length", "i4"),
    ]
)

magic_num = b"constrct"
magic_num_len = len(magic_num)

max_format_version = 0
max_section_version = 0


class ConstructionFormatWrapper(StructureFormatWrapper):
    def __init__(self, path: str):
        super().__init__(path)

        self._format_version: int = max_format_version
        self._section_version: int = max_section_version

        # which sections are in a given chunk
        self._chunk_to_section: Dict[Tuple[int, int], List[ConstructionSection]] = {}

        self._selection_boxes: List[SelectionBox] = []

        # which selection boxes intersect a given chunk (boxes are clipped to the size of the chunk)
        self._chunk_to_box: Dict[Tuple[int, int], List[SelectionBox]] = {}

        self._shallow_load()

    def _shallow_load(self):
        if os.path.isfile(self.path):
            with open(self.path, "rb") as f:
                magic_num_1 = f.read(magic_num_len)
                if magic_num_1 == magic_num:
                    format_version = struct.unpack(">B", f.read(1))[0]
                    if format_version == 0:
                        f.seek(-magic_num_len, os.SEEK_END)
                        magic_num_2 = f.read(magic_num_len)
                        if magic_num_2 == magic_num:
                            f.seek(-magic_num_len - INT_STRUCT.size, os.SEEK_END)
                            metadata_end = f.tell()
                            metadata_start = INT_STRUCT.unpack(f.read(INT_STRUCT.size))[
                                0
                            ]
                            f.seek(metadata_start)

                            metadata = amulet_nbt.load(
                                f.read(metadata_end - metadata_start),
                                compressed=True,
                            )

                            try:
                                platform = metadata["export_version"]["edition"].value
                                version = tuple(
                                    map(
                                        lambda v: v.value,
                                        metadata["export_version"]["version"],
                                    )
                                )
                                selection_boxes = (
                                    metadata["selection_boxes"]
                                    .value.reshape(-1, 6)
                                    .tolist()
                                )
                            except:
                                pass
                            else:
                                self._platform = platform
                                self._version = version
                                self._selection = SelectionGroup(
                                    [
                                        SelectionBox(
                                            (minx, miny, minz), (maxx, maxy, maxz)
                                        )
                                        for minx, miny, minz, maxx, maxy, maxz in selection_boxes
                                    ]
                                )

    def _create(
        self,
        overwrite: bool,
        format_version=max_format_version,
        section_version=max_section_version,
        **kwargs,
    ):
        if not overwrite and os.path.isfile(self.path):
            raise ObjectWriteError(f"There is already a file at {self.path}")
        self._format_version = format_version
        self._section_version = section_version
        translator_version = self.translation_manager.get_version(
            self.platform, self.version
        )
        self._platform = translator_version.platform
        self._version = translator_version.version_number
        self._chunk_to_section = {}
        self._chunk_to_box = {}
        self._populate_chunk_to_box()

    def _populate_chunk_to_box(self):
        for box in self._selection.selection_boxes:
            for cx, cz in box.chunk_locations():
                self._chunk_to_box.setdefault((cx, cz), []).append(
                    SelectionBox.create_chunk_box(cx, cz).intersection(box)
                )

    def open_from(self, f: BinaryIO):
        f = BytesIO(f.read())
        magic_num_1 = f.read(magic_num_len)
        assert magic_num_1 == magic_num, f"This file is not a construction file."
        self._format_version = struct.unpack(">B", f.read(1))[0]
        if self._format_version == 0:
            f.seek(-magic_num_len, os.SEEK_END)
            magic_num_2 = f.read(magic_num_len)
            assert (
                magic_num_2 == magic_num
            ), "It looks like this file is corrupt. It probably wasn't saved properly"

            f.seek(-magic_num_len - INT_STRUCT.size, os.SEEK_END)
            metadata_end = f.tell()
            metadata_start = INT_STRUCT.unpack(f.read(INT_STRUCT.size))[0]
            f.seek(metadata_start)

            metadata = amulet_nbt.load(
                f.read(metadata_end - metadata_start),
                compressed=True,
            )

            try:
                self._platform = metadata["export_version"]["edition"].value
                self._version = tuple(
                    map(lambda v: v.value, metadata["export_version"]["version"])
                )
            except KeyError as e:
                raise KeyError(f'Missing export version identifying key "{e.args[0]}"')

            self._section_version = metadata["section_version"].value

            palette = unpack_palette(metadata["block_palette"])

            selection_boxes = metadata["selection_boxes"].value.reshape(-1, 6).tolist()

            self._selection = SelectionGroup(
                [
                    SelectionBox((minx, miny, minz), (maxx, maxy, maxz))
                    for minx, miny, minz, maxx, maxy, maxz in selection_boxes
                ]
            )

            self._populate_chunk_to_box()

            section_index_table = (
                metadata["section_index_table"].value.view(SECTION_ENTRY_TYPE).tolist()
            )

            if self._section_version == 0:
                for (
                    start_x,
                    start_y,
                    start_z,
                    shape_x,
                    shape_y,
                    shape_z,
                    position,
                    length,
                ) in section_index_table:
                    f.seek(position)
                    nbt_obj = amulet_nbt.load(f.read(length))
                    if nbt_obj["blocks_array_type"].value == -1:
                        blocks = None
                        block_entities = None
                    else:
                        blocks = numpy.reshape(
                            nbt_obj["blocks"].value, (shape_x, shape_y, shape_z)
                        )
                        block_entities = parse_block_entities(nbt_obj["block_entities"])

                    start = numpy.array([start_x, start_y, start_z])
                    chunk_index: numpy.ndarray = start // self.sub_chunk_size
                    shape = numpy.array([shape_x, shape_y, shape_z])
                    if numpy.any(shape <= 0):
                        continue  # skip sections with zero size
                    if numpy.any(
                        start + shape > (chunk_index + 1) * self.sub_chunk_size
                    ):
                        log.error(
                            f"section in construction file did not fit in one sub-chunk. Start: {start}, Shape: {shape}"
                        )
                    cx, cy, cz = chunk_index.tolist()
                    self._chunk_to_section.setdefault((cx, cz), []).append(
                        ConstructionSection(
                            (start_x, start_y, start_z),
                            (shape_x, shape_y, shape_z),
                            blocks,
                            palette,
                            parse_entities(nbt_obj["entities"]),
                            block_entities,
                        )
                    )
            else:
                raise Exception(
                    f"This wrapper does not support any construction section version higher than {max_section_version}"
                )

        else:
            raise Exception(
                f"This wrapper does not support any construction format version higher than {max_format_version}"
            )

    @property
    def multi_selection(self) -> bool:
        return True

    @staticmethod
    def is_valid(path: str) -> bool:
        """
        Returns whether this format is able to load the given object.

        :param path: The path of the object to load.
        :return: True if the world can be loaded by this format, False otherwise.
        """
        return os.path.isfile(path) and path.endswith(".construction")

    @property
    def valid_formats(self) -> Dict[PlatformType, Tuple[bool, bool]]:
        return {
            "bedrock": (True, True),
            "java": (True, True),
        }

    @property
    def extensions(self) -> Tuple[str, ...]:
        return (".construction",)

    def _get_interface(
        self, raw_chunk_data: Optional[Any] = None
    ) -> Construction0Interface:
        return construction_0_interface

    def _get_interface_and_translator(
        self, raw_chunk_data=None
    ) -> Tuple["Interface", "Translator", VersionNumberAny]:
        interface = self._get_interface(raw_chunk_data)
        translator, version_identifier = interface.get_translator(
            self.max_world_version, raw_chunk_data, self.translation_manager
        )
        return interface, translator, version_identifier

    def save_to(self, f: BinaryIO):
        palette: BlockManager = BlockManager()
        f.write(magic_num)
        f.write(struct.pack(">B", self._format_version))
        if self._format_version == 0:
            metadata = amulet_nbt.NBTFile(
                amulet_nbt.TAG_Compound(
                    {
                        "created_with": amulet_nbt.TAG_String(
                            "amulet_python_wrapper_v2"
                        ),
                        "selection_boxes": amulet_nbt.TAG_Int_Array(
                            [
                                c
                                for box in self._selection.selection_boxes
                                for c in (*box.min, *box.max)
                            ]
                        ),
                        "section_version": amulet_nbt.TAG_Byte(self._section_version),
                        "export_version": amulet_nbt.TAG_Compound(
                            {
                                "edition": amulet_nbt.TAG_String(self._platform),
                                "version": amulet_nbt.TAG_List(
                                    [amulet_nbt.TAG_Int(v) for v in self._version]
                                ),
                            }
                        ),
                    }
                )
            )
            section_index_table: List[
                Tuple[int, int, int, int, int, int, int, int]
            ] = []
            if self._section_version == 0:
                for section_list in self._chunk_to_section.values():
                    for section in section_list:
                        sx, sy, sz = section.location
                        shapex, shapey, shapez = section.shape
                        blocks = section.blocks
                        entities = section.entities
                        block_entities = section.block_entities
                        section_palette = section.palette
                        position = f.tell()

                        _tag = amulet_nbt.TAG_Compound(
                            {"entities": serialise_entities(entities)}
                        )

                        if blocks is None:
                            _tag["blocks_array_type"] = amulet_nbt.TAG_Byte(-1)
                        else:
                            flattened_array = blocks.ravel()
                            index, flattened_array = numpy.unique(
                                flattened_array, return_inverse=True
                            )
                            section_palette = numpy.array(
                                section_palette, dtype=object
                            )[index]
                            lut = numpy.vectorize(palette.get_add_block)(
                                section_palette
                            )
                            flattened_array = lut[flattened_array]
                            array_type = find_fitting_array_type(flattened_array)
                            _tag["blocks_array_type"] = amulet_nbt.TAG_Byte(
                                array_type().tag_id
                            )
                            _tag["blocks"] = array_type(flattened_array)
                            _tag["block_entities"] = serialise_block_entities(
                                block_entities or []
                            )

                        amulet_nbt.NBTFile(_tag).save_to(f)

                        length = f.tell() - position
                        section_index_table.append(
                            (sx, sy, sz, shapex, shapey, shapez, position, length)
                        )
            else:
                raise Exception(
                    f"This wrapper doesn't support any section version higher than {max_section_version}"
                )
            metadata_start = f.tell()
            metadata["section_index_table"] = amulet_nbt.TAG_Byte_Array(
                numpy.array(section_index_table, dtype=SECTION_ENTRY_TYPE).view(
                    numpy.int8
                )
            )
            metadata["block_palette"] = pack_palette(palette)
            metadata.save_to(f)
            f.write(INT_STRUCT.pack(metadata_start))
            f.write(magic_num)
        else:
            raise Exception(
                f"This wrapper doesn't support any construction version higher than {max_format_version}"
            )

    def _close(self):
        """Close the disk database"""
        pass

    def unload(self):
        """Unload data stored in the Format class"""
        pass

    def all_chunk_coords(
        self, dimension: Optional[Dimension] = None
    ) -> Generator[Tuple[int, int], None, None]:
        """A generator of all chunk coords"""
        yield from self._chunk_to_section.keys()

    def has_chunk(self, cx: int, cz: int, dimension: Dimension) -> bool:
        return (cx, cz) in self._chunk_to_section

    def _pack(
        self,
        chunk: "Chunk",
        translator: "Translator",
        chunk_version: VersionNumberAny,
    ) -> Tuple["Chunk", AnyNDArray]:
        return chunk, numpy.array(chunk.block_palette.blocks())

    def _encode(
        self,
        chunk: Chunk,
        chunk_palette: AnyNDArray,
        interface: ConstructionInterface,
    ):
        return interface.encode(
            chunk,
            chunk_palette,
            self.max_world_version,
            self._chunk_to_box.get((chunk.cx, chunk.cz)),
        )

    def _unpack(
        self,
        translator: "Translator",
        game_version: VersionNumberAny,
        chunk: "Chunk",
        chunk_palette: AnyNDArray,
    ) -> "Chunk":
        palette = chunk._block_palette = BlockManager()
        lut = numpy.array([palette.get_add_block(block) for block in chunk_palette])
        if len(palette.blocks()) != len(chunk_palette):
            # if a blockstate was defined twice
            for cy in chunk.blocks.sub_chunks:
                chunk.blocks.add_sub_chunk(cy, lut[chunk.blocks.get_sub_chunk(cy)])
        return chunk

    def _delete_chunk(self, cx: int, cz: int, dimension: Optional[Dimension] = None):
        if (cx, cz) in self._chunk_to_section:
            del self._chunk_to_section[(cx, cz)]

    def _put_raw_chunk_data(
        self,
        cx: int,
        cz: int,
        data: List[ConstructionSection],
        dimension: Optional[Dimension] = None,
    ):
        """
        Actually stores the data from the interface to disk.
        """
        self._chunk_to_section[cx, cz] = copy.deepcopy(data)

    def _get_raw_chunk_data(
        self, cx: int, cz: int, dimension: Optional[Dimension] = None
    ) -> List[ConstructionSection]:
        """
        Return the raw data as loaded from disk.

        :param cx: The x coordinate of the chunk.
        :param cz: The z coordinate of the chunk.
        :param dimension: The dimension to load the data from.
        :return: The raw chunk data.
        """
        if (cx, cz) in self._chunk_to_section:
            return copy.deepcopy(self._chunk_to_section[(cx, cz)])
        else:
            raise ChunkDoesNotExist



from __future__ import annotations

from .leveldb_21 import (
    LevelDB21Interface,
)


class LevelDB22Interface(LevelDB21Interface):
    def __init__(self):
        super().__init__()

        self.features["chunk_version"] = 22


export = LevelDB22Interface



from __future__ import annotations

import numpy
from typing import TYPE_CHECKING

from amulet.level.translators.chunk.bedrock import BaseBedrockTranslator
from amulet.api.data_types import BlockNDArray, AnyNDArray
from amulet.api.block import Block

if TYPE_CHECKING:
    from PyMCTranslate import Version


class BedrockNBTBlockstateTranslator(BaseBedrockTranslator):
    @staticmethod
    def is_valid(key):
        return key[0] == "bedrock" and (1, 13, 0) <= key[1]

    def _pack_block_palette(
        self, version: "Version", palette: BlockNDArray
    ) -> AnyNDArray:
        """
        Packs a numpy array of Block objects into an object array containing one more more pairs of version number and Block objects.
        :param version:
        :param palette:
        :return: numpy.ndarray[Tuple[Tuple[Optional[VersionNumber], Block], ...]]
        """
        palette_ = numpy.empty(len(palette), dtype=object)
        for palette_index, block in enumerate(palette):
            block: "Block"
            # TODO: perhaps check that all properties are NBT objects user interaction if not
            blocks = []
            for b in block.block_tuple:
                if "__version__" in b.properties:
                    properties = b.properties
                    version_number = properties.pop("__version__").value
                    b = Block(b.namespace, b.base_name, properties, b.extra_blocks)
                else:
                    version_number = None
                blocks.append((version_number, b))
            palette_[palette_index] = tuple(blocks)

        return palette_


export = BedrockNBTBlockstateTranslator



from __future__ import annotations

from .anvil_1503 import (
    Anvil1503Interface,
)


class Anvil1908Interface(Anvil1503Interface):
    def __init__(self):
        Anvil1503Interface.__init__(self)
        self.features["height_map"] = "C|V4"

    @staticmethod
    def minor_is_valid(key: int):
        return 1908 <= key < 1912


export = Anvil1908Interface



from __future__ import annotations
import warnings

from .base_level import BaseLevel
from amulet.api.wrapper import WorldFormatWrapper


class World(BaseLevel):
    """
    Class that handles world editing of any world format via an separate and flexible data format
    """

    def __init__(
        self, directory: str, world_wrapper: WorldFormatWrapper, temp_dir: str = None
    ):
        assert isinstance(world_wrapper, WorldFormatWrapper)
        super().__init__(directory, world_wrapper, temp_dir)

    @property
    def level_wrapper(self) -> "WorldFormatWrapper":
        """A class to access data directly from the level."""
        return self._level_wrapper

    @property
    def world_wrapper(self) -> WorldFormatWrapper:
        warnings.warn(
            "BaseLevel.world_wrapper is depreciated and will be removed in the future. Please use BaseLevel.level_wrapper instead.",
            DeprecationWarning,
        )
        return self._level_wrapper



from __future__ import annotations

from .anvil_1444 import (
    Anvil1444Interface,
)


class Anvil1466Interface(Anvil1444Interface):
    def __init__(self):
        Anvil1444Interface.__init__(self)
        self.features["height_map"] = "C|V1"

    @staticmethod
    def minor_is_valid(key: int):
        return 1466 <= key < 1467


export = Anvil1466Interface



"""Varint encoder/decoder

varints are a common encoding for variable length integer data, used in
libraries such as sqlite, protobuf, v8, and more.

Here's a quick and dirty module to help avoid reimplementing the same thing
over and over again.

With modifications to allow reading and writing an array of varints.
"""

from io import BytesIO
from typing import List, Iterable


def _read_byte(stream: BytesIO) -> int:
    """Read a byte from the file (as an integer)

    :param stream: The stream to read the byte from.
    :return: The integer value of the byte.
    :raises:
        EOFError: if the stream ends while reading bytes.
    """
    c = stream.read(1)
    if not c:
        raise EOFError("Unexpected EOF while reading bytes")
    return ord(c)


def decode_stream_one(stream: BytesIO) -> int:
    """Read a varint from `stream`"""
    shift = 0
    result = 0
    while True:
        i = _read_byte(stream)
        result |= (i & 0x7F) << shift
        shift += 7
        if not (i & 0x80):
            break
    return result


def decode_bytes_one(buf: bytes):
    """Read a varint from `buf` bytes"""
    return decode_stream_one(BytesIO(buf))


def decode_stream(stream: BytesIO) -> List[int]:
    """Read varint(s) from `stream` until the stream has been exhausted."""
    start = stream.tell()
    end = stream.seek(0, 2)
    stream.seek(start)

    result = []
    while stream.tell() < end:
        result.append(decode_stream_one(stream))
    return result


def decode_bytes(buf: bytes) -> List[int]:
    """Read varint(s) from `buf` bytes until all the bytes have been exhausted."""
    return decode_stream(BytesIO(buf))


def decode_byte_array(buf: Iterable[int]) -> List[int]:
    """Read varint(s) from an iterable of ints until the iterable has been exhausted.
    All ints must be in range(0, 256)"""
    return decode_bytes(bytes(buf))


def encode(number: int):
    """Pack `number` into varint bytes"""
    buf = []
    while True:
        towrite = number & 0x7F
        number >>= 7
        if number:
            buf.append(bytes((towrite | 0x80,)))
        else:
            buf.append(bytes((towrite,)))
            break
    return b"".join(buf)


def encode_array(array: Iterable[int]) -> bytes:
    """Pack an array of integers into a sequence of packed varint(s)."""
    return b"".join(encode(n) for n in array)



from __future__ import annotations

from .leveldb_18 import (
    LevelDB18Interface,
)


class LevelDB19Interface(LevelDB18Interface):
    def __init__(self):
        super().__init__()

        self.features["chunk_version"] = 19


export = LevelDB19Interface



import os
from typing import Optional, Tuple, Generator, TYPE_CHECKING, BinaryIO, Dict, List
import numpy
import copy

import amulet_nbt

from amulet.api.data_types import (
    VersionNumberAny,
    VersionNumberInt,
    ChunkCoordinates,
    AnyNDArray,
    Dimension,
    PlatformType,
)
from amulet.api.wrapper import StructureFormatWrapper
from amulet.api.chunk import Chunk
from amulet.api.selection import SelectionGroup, SelectionBox
from amulet.api.errors import ChunkDoesNotExist, ObjectWriteError, ObjectReadError
from amulet.api.block import Block
from amulet.utils.numpy_helpers import brute_sort_objects_no_hash

from .chunk import SpongeSchemChunk
from .interface import SpongeSchemInterface
from .varint import decode_byte_array, encode_array

if TYPE_CHECKING:
    from amulet.api.wrapper import Translator, Interface


class SpongeSchemReadError(ObjectReadError):
    pass


class SpongeSchemWriteError(ObjectWriteError):
    pass


sponge_schem_interface = SpongeSchemInterface()

max_schem_version = 2


class SpongeSchemFormatWrapper(StructureFormatWrapper):
    _version: VersionNumberInt

    def __init__(self, path: str):
        super().__init__(path)
        self._chunks: Dict[
            ChunkCoordinates,
            Tuple[
                SelectionBox,
                numpy.ndarray,
                AnyNDArray,
                List[amulet_nbt.TAG_Compound],
                List[amulet_nbt.TAG_Compound],
            ],
        ] = {}
        self._schem_version: int = max_schem_version

    def _create(self, overwrite: bool, **kwargs):
        if not overwrite and os.path.isfile(self.path):
            raise SpongeSchemWriteError(f"There is already a file at {self.path}")
        translator_version = self.translation_manager.get_version("java", self._version)
        self._platform = translator_version.platform
        self._version = translator_version.data_version
        self._chunks = {}

    def open_from(self, f: BinaryIO):
        sponge_schem = amulet_nbt.load(f)
        version = sponge_schem.get("Version")
        if not isinstance(version, amulet_nbt.TAG_Int):
            raise SpongeSchemReadError("Version key must exist and be an integer.")
        if version == 1:
            raise SpongeSchemReadError(
                "Sponge Schematic Version 1 is not supported currently."
            )
        elif version == 2:
            offset = sponge_schem.get("Offset")
            if isinstance(offset, amulet_nbt.TAG_Int_Array) and len(offset) == 3:
                min_point = numpy.array(offset)
            else:
                min_point = numpy.array([0, 0, 0], dtype=numpy.int32)

            size = []
            for key in ("Width", "Height", "Length"):
                val = sponge_schem.get(key)
                if not isinstance(val, amulet_nbt.TAG_Short):
                    raise SpongeSchemReadError(
                        f"Key {key} must exist and be a TAG_Short."
                    )
                # convert to an unsigned short
                val = val.value
                if val < 0:
                    val += 2 ** 16
                size.append(val)

            max_point = min_point + size
            selection = SelectionBox(min_point, max_point)
            self._selection = SelectionGroup(selection)
            data_version = sponge_schem.get("DataVersion")
            if not isinstance(data_version, amulet_nbt.TAG_Int):
                raise SpongeSchemReadError("DataVersion must be a TAG_Int.")
            translator_version = self.translation_manager.get_version(
                "java", int(data_version)
            )
            self._platform = translator_version.platform
            self._version = translator_version.data_version

            packed_block_data = sponge_schem.get("BlockData")
            if not isinstance(packed_block_data, amulet_nbt.TAG_Byte_Array):
                raise SpongeSchemReadError("BlockData must be a TAG_Byte_Array")

            unpacked_block_data = decode_byte_array(
                numpy.array(packed_block_data, dtype=numpy.uint8)
            )
            if len(unpacked_block_data) != numpy.prod(size):
                raise SpongeSchemReadError(
                    "The data contained in BlockData does not match the size of the schematic."
                )
            dx, dy, dz = selection.shape
            blocks_array: numpy.ndarray = numpy.transpose(
                numpy.array(
                    unpacked_block_data,
                    dtype=numpy.uint32,
                ).reshape((dy, dz, dx)),
                (2, 0, 1),  # YZX => XYZ
            )

            if "Palette" not in sponge_schem:
                raise SpongeSchemReadError(
                    "Amulet is not able to read Sponge Schem files with no block palette."
                )

            palette_data = sponge_schem.get("Palette")
            if not isinstance(palette_data, amulet_nbt.TAG_Compound):
                raise SpongeSchemReadError("Palette must be a TAG_Compound.")

            block_palette: Dict[int, Block] = {}
            for blockstate, index in palette_data.items():
                if index.value in block_palette:
                    raise SpongeSchemReadError(
                        f"Duplicate block index {index} found in the palette."
                    )
                block_palette[index.value] = Block.from_string_blockstate(blockstate)

            if not numpy.all(numpy.isin(blocks_array, list(block_palette))):
                raise SpongeSchemReadError(
                    "Some values in BlockData were not present in Palette"
                )

            for cx, cz in selection.chunk_locations():
                chunk_box = SelectionBox.create_chunk_box(cx, cz).intersection(
                    selection
                )
                array_slice = chunk_box.create_moved_box(
                    selection.min, subtract=True
                ).slice
                chunk_blocks_: numpy.ndarray = blocks_array[array_slice]
                chunk_palette_indexes, chunk_blocks = numpy.unique(
                    chunk_blocks_,
                    return_inverse=True,
                )
                chunk_blocks = chunk_blocks.reshape(chunk_blocks_.shape)

                chunk_palette = numpy.empty(len(chunk_palette_indexes), dtype=object)
                for palette_index, index in enumerate(chunk_palette_indexes):
                    chunk_palette[palette_index] = block_palette[index]

                self._chunks[(cx, cz)] = (
                    chunk_box,
                    chunk_blocks,
                    chunk_palette,
                    [],
                    [],
                )

            if "BlockEntities" in sponge_schem:
                block_entities = sponge_schem["BlockEntities"]
                if (
                    not isinstance(block_entities, amulet_nbt.TAG_List)
                    or block_entities.list_data_type
                    != 10  # amulet_nbt.TAG_Compound.tag_id
                ):
                    raise SpongeSchemReadError(
                        "BlockEntities must be a TAG_List of compound tags."
                    )

                for block_entity in block_entities:
                    if "Pos" in block_entity:
                        pos = block_entity["Pos"]
                        if isinstance(pos, amulet_nbt.TAG_Int_Array) and len(pos) == 3:
                            pos = pos + min_point
                            x, y, z = (
                                pos[0],
                                pos[1],
                                pos[2],
                            )
                            block_entity["Pos"] = amulet_nbt.TAG_Int_Array(pos)
                            cx, cz = x >> 4, z >> 4
                            if (cx, cz) in self._chunks and (x, y, z) in self._chunks[
                                (cx, cz)
                            ][0]:
                                self._chunks[(cx, cz)][3].append(block_entity)

            if "Entities" in sponge_schem:
                entities = sponge_schem["Entities"]
                if (
                    not isinstance(entities, amulet_nbt.TAG_List)
                    or entities.list_data_type != 10  # amulet_nbt.TAG_Compound.tag_id
                ):
                    raise SpongeSchemReadError(
                        "Entities must be a TAG_List of compound tags."
                    )

                for entity in entities:
                    if "Pos" in entity:
                        pos = entity["Pos"]
                        if (
                            isinstance(pos, amulet_nbt.TAG_List)
                            and len(pos) == 3
                            and pos.list_data_type == 6
                        ):  # amulet_nbt.TAG_Double.tag_id:
                            x, y, z = (
                                pos[0].value + offset[0],
                                pos[1].value + offset[0],
                                pos[2].value + offset[0],
                            )
                            entity["Pos"] = amulet_nbt.TAG_List(
                                [
                                    amulet_nbt.TAG_Int(x),
                                    amulet_nbt.TAG_Int(y),
                                    amulet_nbt.TAG_Int(z),
                                ]
                            )
                            cx, cz = numpy.floor([x, z]).astype(int) >> 4
                            if (cx, cz) in self._chunks and (x, y, z) in self._chunks[
                                (cx, cz)
                            ][0]:
                                self._chunks[(cx, cz)][4].append(entity)

        else:
            raise SpongeSchemReadError(
                f"Sponge Schematic Version {version.value} is not supported currently."
            )

    @staticmethod
    def is_valid(path: str) -> bool:
        """
        Returns whether this format is able to load the given object.

        :param path: The path of the object to load.
        :return: True if the world can be loaded by this format, False otherwise.
        """
        return os.path.isfile(path) and path.endswith(".schem")

    @property
    def valid_formats(self) -> Dict[PlatformType, Tuple[bool, bool]]:
        return {"java": (False, True)}

    @property
    def extensions(self) -> Tuple[str, ...]:
        return (".schem",)

    def _get_interface(self, raw_chunk_data=None) -> "SpongeSchemInterface":
        return sponge_schem_interface

    def _get_interface_and_translator(
        self, raw_chunk_data=None
    ) -> Tuple["Interface", "Translator", VersionNumberAny]:
        interface = self._get_interface(raw_chunk_data)
        translator, version_identifier = interface.get_translator(
            self.max_world_version, raw_chunk_data, self.translation_manager
        )
        return interface, translator, version_identifier

    def save_to(self, f: BinaryIO):
        if self._schem_version == 1:
            raise SpongeSchemReadError(
                "Sponge Schematic Version 1 is not supported currently."
            )
        elif self._schem_version == 2:
            selection = self._selection.selection_boxes[0]
            if any(s > 2 ** 16 - 1 for s in selection.shape):
                raise SpongeSchemWriteError(
                    "The structure is too large to be exported to a Sponge Schematic file. It must be 2^16 - 1 at most in each dimension."
                )
            overflowed_shape = [
                s if s < 2 ** 15 else s - 2 ** 16 for s in selection.shape
            ]
            data = amulet_nbt.NBTFile(
                amulet_nbt.TAG_Compound(
                    {
                        "Version": amulet_nbt.TAG_Int(2),
                        "DataVersion": amulet_nbt.TAG_Int(self._version),
                        "Width": amulet_nbt.TAG_Short(overflowed_shape[0]),
                        "Height": amulet_nbt.TAG_Short(overflowed_shape[1]),
                        "Length": amulet_nbt.TAG_Short(overflowed_shape[2]),
                        "Offset": amulet_nbt.TAG_Int_Array(selection.min),
                    }
                )
            )

            entities = []
            block_entities = []
            blocks = numpy.zeros(selection.shape, dtype=numpy.uint32)
            palette: List[AnyNDArray] = []
            if self._version < 1500:
                raise Exception(
                    "Writing to Sponge Schematic files in pre-1.13 format is not currently supported."
                )
            else:
                arr = numpy.empty(1, dtype=object)
                arr[0] = Block("minecraft", "air")
                palette.append((arr))

            palette_len = 1

            for (
                selection_,
                blocks_,
                palette_,
                block_entities_,
                entities_,
            ) in self._chunks.values():
                if selection_.intersects(selection):
                    box = selection_.create_moved_box(selection.min, subtract=True)
                    blocks[box.slice] = blocks_ + palette_len
                    palette.append(palette_)
                    palette_len += len(palette_)
                    for be in block_entities_:
                        be = copy.deepcopy(be)
                        be["Pos"] = amulet_nbt.TAG_Int_Array(be["Pos"] - selection.min)
                        block_entities.append(be)

                    for e in entities_:
                        e = copy.deepcopy(e)
                        x, y, z = e["Pos"]
                        e["Pos"] = amulet_nbt.TAG_List(
                            [
                                amulet_nbt.TAG_Int(x - selection.min_x),
                                amulet_nbt.TAG_Int(y - selection.min_y),
                                amulet_nbt.TAG_Int(z - selection.min_z),
                            ]
                        )
                        entities.append(e)

            compact_palette, lut = brute_sort_objects_no_hash(
                numpy.concatenate(palette)
            )
            blocks = numpy.transpose(lut[blocks], (1, 2, 0)).ravel()  # XYZ => YZX
            block_palette = []
            for index, block in enumerate(compact_palette):
                block: Block
                block_palette.append(block.blockstate)

            data["PaletteMax"] = amulet_nbt.TAG_Int(len(compact_palette))
            data["Palette"] = amulet_nbt.TAG_Compound(
                {
                    blockstate: amulet_nbt.TAG_Int(index)
                    for index, blockstate in enumerate(block_palette)
                }
            )
            data["BlockData"] = amulet_nbt.TAG_Byte_Array(list(encode_array(blocks)))
            if block_entities:
                data["BlockEntities"] = amulet_nbt.TAG_List(block_entities)
            if entities:
                data["Entities"] = amulet_nbt.TAG_List(entities)

            data.save_to(f)
        else:
            raise SpongeSchemReadError(
                f"Sponge Schematic Version {self._schem_version} is not supported currently."
            )

    def _close(self):
        """Close the disk database"""
        self._chunks.clear()

    def unload(self):
        """Unload data stored in the Format class"""
        pass

    def all_chunk_coords(
        self, dimension: Optional[Dimension] = None
    ) -> Generator[ChunkCoordinates, None, None]:
        """A generator of all chunk coords"""
        yield from self._chunks.keys()

    def has_chunk(self, cx: int, cz: int, dimension: Dimension) -> bool:
        return (cx, cz) in self._chunks

    def _encode(
        self,
        chunk: Chunk,
        chunk_palette: AnyNDArray,
        interface: SpongeSchemInterface,
    ):
        return interface.encode(
            chunk,
            chunk_palette,
            self.max_world_version,
            SelectionBox.create_chunk_box(chunk.cx, chunk.cz).intersection(
                self._selection.to_box()
            ),
        )

    def _delete_chunk(self, cx: int, cz: int, dimension: Optional[Dimension] = None):
        if (cx, cz) in self._chunks:
            del self._chunks[(cx, cz)]

    def _put_raw_chunk_data(
        self,
        cx: int,
        cz: int,
        section: SpongeSchemChunk,
        dimension: Optional[Dimension] = None,
    ):
        """
        Actually stores the data from the interface to disk.
        """
        self._chunks[(cx, cz)] = copy.deepcopy(
            (
                section.selection,
                section.blocks,
                section.palette,
                section.block_entities,
                section.entities,
            )
        )

    def _get_raw_chunk_data(
        self, cx: int, cz: int, dimension: Optional[Dimension] = None
    ) -> SpongeSchemChunk:
        """
        Return the raw data as loaded from disk.

        :param cx: The x coordinate of the chunk.
        :param cz: The z coordinate of the chunk.
        :param dimension: The dimension to load the data from.
        :return: The raw chunk data.
        """
        if (cx, cz) in self._chunks:
            return SpongeSchemChunk(*copy.deepcopy(self._chunks[(cx, cz)]))
        else:
            raise ChunkDoesNotExist



from __future__ import annotations

from .anvil_1908 import (
    Anvil1908Interface,
)


class Anvil1912Interface(Anvil1908Interface):
    def __init__(self):
        Anvil1908Interface.__init__(self)
        self.features["status"] = "j14"

    @staticmethod
    def minor_is_valid(key: int):
        return 1912 <= key < 1934


export = Anvil1912Interface



from __future__ import annotations

import os
import struct
from typing import (
    Dict,
    Set,
    Union,
    Optional,
    List,
    TYPE_CHECKING,
)

from amulet.api.errors import ChunkDoesNotExist, LevelDoesNotExist
from amulet.api.data_types import ChunkCoordinates
from amulet.libs.leveldb import LevelDB

if TYPE_CHECKING:
    from amulet.api.data_types import Dimension

InternalDimension = Optional[int]


class LevelDBDimensionManager:
    # tag_ids = {45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 118}

    def __init__(self, directory: str):
        self._directory = directory
        self._db = LevelDB(os.path.join(self._directory, "db"))
        # self._levels format Dict[level, Dict[Tuple[cx, cz], List[Tuple[full_key, key_extension]]]]
        self._levels: Dict[InternalDimension, Set[ChunkCoordinates]] = {}
        self._dimension_name_map: Dict["Dimension", InternalDimension] = {}
        self._batch_temp: Dict[bytes, Union[bytes, None]] = {}

        self.register_dimension(None, "overworld")
        self.register_dimension(1, "nether")
        self.register_dimension(2, "end")

        for key in self._db.keys():
            if 9 <= len(key) <= 10 and key[8] in [44, 118]:  # "," "v"
                self._add_chunk(key)

            elif 13 <= len(key) <= 14 and key[12] in [44, 118]:  # "," "v"
                self._add_chunk(key, has_level=True)

    def save(self):
        batch = {}
        for key, val in self._batch_temp.items():
            if val is None:
                self._db.delete(key)
            else:
                batch[key] = val
        self._db.putBatch(batch)
        self._batch_temp.clear()

    def close(self):
        self._db.close()

    @property
    def dimensions(self) -> List["Dimension"]:
        """A list of all the levels contained in the world"""
        return list(self._dimension_name_map.keys())

    def register_dimension(
        self,
        dimension_internal: InternalDimension,
        dimension_name: Optional["Dimension"] = None,
    ):
        """
        Register a new dimension.
        :param dimension_internal: The internal representation of the dimension
        :param dimension_name: The name of the dimension shown to the user
        :return:
        """
        if dimension_name is None:
            dimension_name: "Dimension" = f"DIM{dimension_internal}"

        if (
            dimension_internal not in self._levels
            and dimension_name not in self._dimension_name_map
        ):
            self._levels[dimension_internal] = set()
            self._dimension_name_map[dimension_name] = dimension_internal

    def _get_internal_dimension(self, dimension: "Dimension") -> InternalDimension:
        if dimension in self._dimension_name_map:
            return self._dimension_name_map[dimension]
        else:
            raise LevelDoesNotExist(dimension)

    def all_chunk_coords(self, dimension: "Dimension") -> Set[ChunkCoordinates]:
        internal_dimension = self._get_internal_dimension(dimension)
        if internal_dimension in self._levels:
            return self._levels[internal_dimension]
        else:
            return set()

    def has_chunk(self, cx: int, cz: int, dimension: Dimension) -> bool:
        internal_dimension = self._get_internal_dimension(dimension)
        return (
            internal_dimension in self._levels
            and (cx, cz) in self._levels[internal_dimension]
        )

    def _add_chunk(self, key_: bytes, has_level: bool = False):
        if has_level:
            cx, cz, level = struct.unpack("<iii", key_[:12])
        else:
            cx, cz = struct.unpack("<ii", key_[:8])
            level = None
        if level not in self._levels:
            self.register_dimension(level)
        self._levels[level].add((cx, cz))

    def get_chunk_data(
        self, cx: int, cz: int, dimension: "Dimension"
    ) -> Dict[bytes, bytes]:
        """Get a dictionary of chunk key extension in bytes to the raw data in the key.
        chunk key extension are the character(s) after <cx><cz>[level] in the key
        Will raise ChunkDoesNotExist if the chunk does not exist
        """
        iter_start = struct.pack("<ii", cx, cz)
        iter_end = iter_start + b"\xff"
        internal_dimension = self._get_internal_dimension(dimension)
        if (
            internal_dimension in self._levels
            and (cx, cz) in self._levels[internal_dimension]
        ):
            chunk_data = {}
            for key, val in self._db.iterate(iter_start, iter_end):
                if internal_dimension is None:
                    if 9 <= len(key) <= 10:
                        key_extension = key[8:]
                    else:
                        continue
                else:
                    if (
                        13 <= len(key) <= 14
                        and struct.unpack("<i", key[8:12])[0] == internal_dimension
                    ):
                        key_extension = key[12:]
                    else:
                        continue
                chunk_data[key_extension] = val
            return chunk_data
        else:
            raise ChunkDoesNotExist

    def put_chunk_data(
        self, cx: int, cz: int, data: Dict[bytes, bytes], dimension: "Dimension"
    ):
        """pass data to the region file class"""
        # get the region key
        internal_dimension = self._get_internal_dimension(dimension)
        self._levels[internal_dimension].add((cx, cz))
        if internal_dimension is None:
            key_prefix = struct.pack("<ii", cx, cz)
        else:
            key_prefix = struct.pack("<iii", cx, cz, internal_dimension)
        for key, val in data.items():
            self._batch_temp[key_prefix + key] = val

    def delete_chunk(self, cx: int, cz: int, dimension: "Dimension"):
        if dimension in self._dimension_name_map:
            internal_dimension = self._dimension_name_map[dimension]
            if (
                internal_dimension in self._levels
                and (cx, cz) in self._levels[internal_dimension]
            ):
                chunk_data = self.get_chunk_data(cx, cz, dimension)
                self._levels[internal_dimension].remove((cx, cz))
                for key in chunk_data.keys():
                    if internal_dimension is None:
                        key_prefix = struct.pack("<ii", cx, cz)
                    else:
                        key_prefix = struct.pack("<iii", cx, cz, internal_dimension)

                    self._batch_temp[key_prefix + key] = None



import math
import numpy
from amulet.api.data_types import FloatTriplet, PointCoordinates


def scale_matrix(sx: float, sy: float, sz: float) -> numpy.ndarray:
    """Create a scale matrix from the inputs specified

    :param sx: The scale in the x axis
    :param sy: The scale in the y axis
    :param sz: The scale in the z axis
    :return: The 4x4 scale matrix
    """
    return numpy.array(
        [[sx, 0, 0, 0], [0, sy, 0, 0], [0, 0, sz, 0], [0, 0, 0, 1]], dtype=numpy.float64
    )


def displacement_matrix(x: float, y: float, z: float) -> numpy.ndarray:
    """Create a displacement matrix from the inputs specified

    :param x: The displacement in the x axis
    :param y: The displacement in the y axis
    :param z: The displacement in the z axis
    :return: The 4x4 displacement matrix
    """
    return numpy.array(
        [[1, 0, 0, x], [0, 1, 0, y], [0, 0, 1, z], [0, 0, 0, 1]], dtype=numpy.float64
    )


def _rotation_matrix(*angles: float, order=None) -> numpy.ndarray:
    """Create a rotation matrix from the inputs specified

    :param angles: The angles in radians
    :param order: The order the angles are specified. Transforms will be applied in this order.
    :return: The 4x4 rotation matrix
    """
    assert isinstance(order, str) and len(order) == len(
        angles
    ), "Order must be a string of the same length as angles."
    mat = numpy.identity(4, dtype=numpy.float64)

    for angle, axis in zip(angles, order):
        if angle:
            c = math.cos(angle)
            s = math.sin(angle)
            if axis == "x":
                mat = numpy.matmul(
                    numpy.array(
                        [[1, 0, 0, 0], [0, c, -s, 0], [0, s, c, 0], [0, 0, 0, 1]],
                        dtype=numpy.float64,
                    ),
                    mat,
                )
            elif axis == "y":
                mat = numpy.matmul(
                    numpy.array(
                        [[c, 0, s, 0], [0, 1, 0, 0], [-s, 0, c, 0], [0, 0, 0, 1]],
                        dtype=numpy.float64,
                    ),
                    mat,
                )
            elif axis == "z":
                mat = numpy.matmul(
                    numpy.array(
                        [[c, -s, 0, 0], [s, c, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]],
                        dtype=numpy.float64,
                    ),
                    mat,
                )

    return mat


def rotation_matrix_x(rx: float) -> numpy.ndarray:
    """Create a rotation matrix in the x axis

    :param rx: The angle in radians
    :return: The 4x4 rotation matrix
    """
    return _rotation_matrix(rx, order="x")


def rotation_matrix_y(ry: float) -> numpy.ndarray:
    """Create a rotation matrix in the x axis

    :param ry: The angle in radians
    :return: The 4x4 rotation matrix
    """
    return _rotation_matrix(ry, order="y")


def rotation_matrix_z(rz: float) -> numpy.ndarray:
    """Create a rotation matrix in the x axis

    :param rz: The angle in radians
    :return: The 4x4 rotation matrix
    """
    return _rotation_matrix(rz, order="z")


def rotation_matrix_xy(rx: float, ry: float) -> numpy.ndarray:
    """Create a rotation matrix from the inputs specified

    :param rx: The rotation in radians in the x axis
    :param ry: The rotation in radians in the y axis
    :return: The 4x4 rotation matrix
    """
    return _rotation_matrix(rx, ry, order="xy")


def rotation_matrix_yx(ry: float, rx: float) -> numpy.ndarray:
    """Create a rotation matrix from the inputs specified

    :param rx: The rotation in radians in the x axis
    :param ry: The rotation in radians in the y axis
    :return: The 4x4 rotation matrix
    """
    return _rotation_matrix(ry, rx, order="yx")


def rotation_matrix_xyz(x: float, y: float, z: float) -> numpy.ndarray:
    """Create a rotation matrix from the inputs specified

    :param x: The rotation in radians in the x axis
    :param y: The rotation in radians in the y axis
    :param z: The rotation in radians in the z axis
    :return: The 4x4 rotation matrix
    """
    return _rotation_matrix(x, y, z, order="xyz")


def transform_matrix(
    scale: FloatTriplet,
    rotation: FloatTriplet,
    displacement: PointCoordinates,
    order="xyz",
):
    """Create a 4x4 transformation matrix from the scale, rotation and displacement specified.

    :param scale: The scale in the x, y and z axis
    :param rotation: The rotation in the x, y and z axis (axis can be changed using `order`)
    :param displacement: The displacement in the x, y and z axis
    :param order: The order to apply the rotations in.
    :return: The 4x4 transformation matrix of combined scale, rotation and displacement
    """
    scale_transform = scale_matrix(*scale)
    rotation_transform = _rotation_matrix(*rotation, order=order)
    displacement_transform = displacement_matrix(*displacement)
    return numpy.matmul(
        displacement_transform,
        numpy.matmul(rotation_transform, scale_transform),
    )


def inverse_transform_matrix(
    scale: FloatTriplet,
    rotation: FloatTriplet,
    displacement: PointCoordinates,
    order="xyz",
):
    """Create the inverse of the 4x4 transformation matrix from the scale, rotation and displacement specified.
    This should be the inverse of transform_matrix
    :param scale: The scale in the x, y and z axis
    :param rotation: The rotation in the x, y and z axis (axis can be changed using `order`)
    :param displacement: The displacement in the x, y and z axis
    :param order: The order to apply the rotations in.
    :return: The 4x4 transformation matrix of combined scale, rotation and displacement
    """
    scale_transform = scale_matrix(*1 / numpy.asarray(scale))
    ra, rb, rc = -numpy.asarray(rotation)
    rotation_transform = _rotation_matrix(
        rc, rb, ra, order="".join(list(reversed(order)))
    )
    displacement_transform = displacement_matrix(*-numpy.asarray(displacement))
    return numpy.matmul(
        scale_transform,
        numpy.matmul(rotation_transform, displacement_transform),
    )



from typing import Tuple, Any, Dict, Generator
import threading

from amulet.api.history.data_types import EntryKeyType, EntryType
from amulet.api.history.base import RevisionManager
from .container import ContainerHistoryManager
from amulet.api.errors import EntryDoesNotExist, EntryLoadError
from ..revision_manager import RAMRevisionManager

SnapshotType = Tuple[Any, ...]


class DatabaseHistoryManager(ContainerHistoryManager):
    """Manage the history of a number of items in a database."""

    DoesNotExistError = EntryDoesNotExist
    LoadError = EntryLoadError

    def __init__(self):
        super().__init__()
        self._lock = threading.RLock()
        # this is the database that entries will be directly edited in
        self._temporary_database: Dict[EntryKeyType, EntryType] = {}

        # this is the database where revisions will be cached
        self._history_database: Dict[EntryKeyType, RevisionManager] = {}

    def _check_snapshot(self, snapshot: SnapshotType):
        assert isinstance(snapshot, tuple)

    @property
    def changed(self) -> bool:
        """Have there been modifications since the last save."""
        if super().changed:
            return True
        try:
            next(self.changed_entries())
        except StopIteration:
            return False
        else:
            return True

    def changed_entries(self) -> Generator[EntryKeyType, None, None]:
        """A generator of all the entry keys that have changed since the last save."""
        changed = set()
        with self._lock:
            for key, entry in self._temporary_database.items():
                if entry is None:
                    # If the temporary entry is deleted but there was no historical
                    # record or the historical record was not deleted
                    if (
                        key not in self._history_database
                        or not self._history_database[key].is_deleted
                    ):
                        changed.add(key)
                        yield key

                elif entry.changed:
                    changed.add(key)
                    yield key
        for key, history_entry in self._history_database.items():
            if history_entry.changed and key not in changed:
                yield key

    def _has_entry(self, key: EntryKeyType):
        """Does the entry exist in one of the databases.
        Subclasses should implement a proper method calling this."""
        return key in self._temporary_database or key in self._history_database

    def _get_entry(self, key: EntryKeyType) -> EntryType:
        """Get a key from the database.
        Subclasses should implement a proper method calling this."""
        with self._lock:
            if key in self._temporary_database:
                entry = self._temporary_database[key]
            else:
                if key in self._temporary_database:
                    entry = self._temporary_database[key]
                elif key in self._history_database:
                    entry = self._temporary_database[key] = self._history_database[
                        key
                    ].get_current_entry()
                else:
                    entry = self._temporary_database[
                        key
                    ] = self._get_register_original_entry(key)
        if entry is None:
            raise self.DoesNotExistError
        return entry

    def _get_register_original_entry(self, key: EntryKeyType) -> EntryType:
        """Get and register the original entry."""
        if key in self._history_database:
            raise Exception(f"The entry for {key} has already been registered.")
        try:
            entry = self._get_entry_from_world(key)
        except EntryDoesNotExist:
            entry = None
        self._history_database[key] = self._create_new_revision_manager(key, entry)
        return entry

    def _get_entry_from_world(self, key: EntryKeyType) -> EntryType:
        """If the entry was not found in the database request it from the world."""
        raise NotImplementedError

    @staticmethod
    def _create_new_revision_manager(
        key: EntryKeyType, original_entry: EntryType
    ) -> RevisionManager:
        """Create an RevisionManager as desired and populate it with the original entry."""
        return RAMRevisionManager(original_entry)

    def _put_entry(self, key: EntryKeyType, entry: EntryType):
        with self._lock:
            entry.changed = True
            self._temporary_database[key] = entry

    def _delete_entry(self, key: EntryKeyType):
        with self._lock:
            self._temporary_database[key] = None

    def create_undo_point(self) -> bool:
        """
        Find all entries in the temporary database that have changed since the last undo point and create a new undo point.
        :return: Was an undo point created. If there were no changes no snapshot will be created.
        """
        with self._lock:
            snapshot = []
            for key, entry in tuple(self._temporary_database.items()):
                if entry is None or entry.changed:
                    if key not in self._history_database:
                        # The entry was added without populating from the world
                        # populate the history with the original entry
                        self._get_register_original_entry(key)
                    history_entry = self._history_database[key]
                    if (entry is None and not history_entry.is_deleted) or (
                        entry is not None and entry.changed
                    ):
                        # if the entry has been modified since the last history version
                        history_entry.put_new_entry(entry)
                        snapshot.append(key)

            self._temporary_database.clear()  # unload all the data from the temporary database
            # so that it is repopulated from the history database. This fixes the issue of entries
            # being modified without the `changed` flag being set to True.

        return self._register_snapshot(tuple(snapshot))

    def _mark_saved(self):
        """Let the class know that the current state has been saved."""
        for entry in self._history_database.values():
            entry.mark_saved()

    def _undo(self, snapshot: SnapshotType):
        """Undoes the last set of changes to the database"""
        with self._lock:
            for key in snapshot:
                self._history_database[key].undo()
                if key in self._temporary_database:
                    del self._temporary_database[key]

    def _redo(self, snapshot: SnapshotType):
        """Redoes the last set of changes to the database"""
        with self._lock:
            for key in snapshot:
                self._history_database[key].redo()
                if key in self._temporary_database:
                    del self._temporary_database[key]

    def restore_last_undo_point(self):
        with self._lock:
            self._temporary_database.clear()

    def purge(self):
        """Unload all cached data. Effectively returns the class to its starting state."""
        with self._lock:
            super().purge()
            self._temporary_database.clear()
            self._history_database.clear()



import os

from .api import *
from .utils.log import log

entity_support = False
IMG_DIRECTORY = os.path.join(os.path.dirname(__file__), "img")

from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions

from amulet.level.load import load_level, load_format



from __future__ import annotations

import numpy
from typing import Tuple, Any, Union, TYPE_CHECKING, Optional, overload, Type
from enum import Enum

from amulet.api.block_entity import BlockEntity
from amulet.api.entity import Entity
from amulet.api.data_types import AnyNDArray
import amulet_nbt
from amulet_nbt import (
    TAG_List,
    TAG_Compound,
    AnyNBT,
)

if TYPE_CHECKING:
    from amulet.api.wrapper import Translator
    from amulet.api.chunk import Chunk


class EntityIDType(Enum):
    int_id = 0
    str_id = 1
    namespace_str_id = 2
    namespace_str_Id = 3
    namespace_str_identifier = 4


class EntityCoordType(Enum):
    xyz_int = 0
    Pos_list_float = 1
    Pos_list_double = 2
    Pos_list_int = 3
    Pos_array_int = 4


PosTypeMap = {
    3: EntityCoordType.Pos_list_int,
    5: EntityCoordType.Pos_list_float,
    6: EntityCoordType.Pos_list_double,
}


class Interface:
    def decode(self, cx: int, cz: int, data: Any) -> Tuple["Chunk", AnyNDArray]:
        """
        Create an amulet.api.chunk.Chunk object from raw data given by the format
        :param cx: chunk x coordinate
        :type cx: int
        :param cz: chunk z coordinate
        :type cz: int
        :param data: Raw chunk data provided by the format.
        :type data: Any
        :return: Chunk object in version-specific format, along with the block_palette for that chunk.
        :rtype: Chunk
        """
        raise NotImplementedError()

    def _decode_entity(
        self,
        nbt: amulet_nbt.NBTFile,
        id_type: EntityIDType,
        coord_type: EntityCoordType,
    ) -> Optional[Entity]:
        entity = self._decode_base_entity(nbt, id_type, coord_type)
        if entity is not None:
            namespace, base_name, x, y, z, nbt = entity
            return Entity(
                namespace=namespace, base_name=base_name, x=x, y=y, z=z, nbt=nbt
            )

    def _decode_block_entity(
        self,
        nbt: amulet_nbt.NBTFile,
        id_type: EntityIDType,
        coord_type: EntityCoordType,
    ) -> Optional[BlockEntity]:
        entity = self._decode_base_entity(nbt, id_type, coord_type)
        if entity is not None:
            namespace, base_name, x, y, z, nbt = entity
            return BlockEntity(
                namespace=namespace, base_name=base_name, x=x, y=y, z=z, nbt=nbt
            )

    @staticmethod
    def _decode_base_entity(
        nbt: amulet_nbt.NBTFile, id_type: EntityIDType, coord_type: EntityCoordType
    ) -> Optional[
        Tuple[
            str,
            str,
            Union[int, float],
            Union[int, float],
            Union[int, float],
            amulet_nbt.NBTFile,
        ]
    ]:
        if not isinstance(nbt, amulet_nbt.NBTFile) and isinstance(
            nbt.value, amulet_nbt.TAG_Compound
        ):
            return

        if id_type == EntityIDType.namespace_str_id:
            entity_id = nbt.pop("id", amulet_nbt.TAG_String(""))
            if (
                not isinstance(entity_id, amulet_nbt.TAG_String)
                or entity_id.value == ""
                or ":" not in entity_id.value
            ):
                return
            namespace, base_name = entity_id.value.split(":", 1)

        elif id_type == EntityIDType.namespace_str_Id:
            entity_id = nbt.pop("Id", amulet_nbt.TAG_String(""))
            if (
                not isinstance(entity_id, amulet_nbt.TAG_String)
                or entity_id.value == ""
                or ":" not in entity_id.value
            ):
                return
            namespace, base_name = entity_id.value.split(":", 1)

        elif id_type == EntityIDType.str_id:
            entity_id = nbt.pop("id", amulet_nbt.TAG_String(""))
            if (
                not isinstance(entity_id, amulet_nbt.TAG_String)
                or entity_id.value == ""
            ):
                return
            namespace = ""
            base_name = entity_id.value

        elif id_type in [EntityIDType.namespace_str_identifier, EntityIDType.int_id]:
            if "identifier" in nbt:
                entity_id = nbt.pop("identifier")
                if (
                    not isinstance(entity_id, amulet_nbt.TAG_String)
                    or entity_id.value == ""
                    or ":" not in entity_id.value
                ):
                    return
                namespace, base_name = entity_id.value.split(":", 1)
            elif "id" in nbt:
                entity_id = nbt.pop("id")
                if not isinstance(entity_id, amulet_nbt.TAG_Int):
                    return
                namespace = ""
                base_name = str(entity_id.value)
            else:
                return
        else:
            raise NotImplementedError(f"Entity id type {id_type}")

        if coord_type in [
            EntityCoordType.Pos_list_double,
            EntityCoordType.Pos_list_float,
            EntityCoordType.Pos_list_int,
        ]:
            if "Pos" not in nbt:
                return
            pos = nbt.pop("Pos")
            pos: amulet_nbt.TAG_List

            if (
                not isinstance(pos, amulet_nbt.TAG_List)
                or len(pos) != 3
                or PosTypeMap.get(pos.list_data_type) != coord_type
            ):
                return
            x, y, z = [c.value for c in pos]
        elif coord_type == EntityCoordType.Pos_array_int:
            if "Pos" not in nbt:
                return
            pos = nbt.pop("Pos")
            pos: amulet_nbt.TAG_Int_Array

            if not isinstance(pos, amulet_nbt.TAG_Int_Array) or len(pos) != 3:
                return
            x, y, z = pos
        elif coord_type == EntityCoordType.xyz_int:
            if not all(
                c in nbt and isinstance(nbt[c], amulet_nbt.TAG_Int)
                for c in ("x", "y", "z")
            ):
                return
            x, y, z = [nbt.pop(c).value for c in ("x", "y", "z")]
        else:
            raise NotImplementedError(f"Entity coord type {coord_type}")

        return namespace, base_name, x, y, z, nbt

    def encode(
        self,
        chunk: "Chunk",
        palette: AnyNDArray,
        max_world_version: Tuple[str, Union[int, Tuple[int, int, int]]],
    ) -> Any:
        """
        Take a version-specific chunk and encode it to raw data for the format to store.
        :param chunk: The already translated version-specfic chunk to encode.
        :type chunk: Chunk
        :param palette: The block_palette the ids in the chunk correspond to.
        :type palette: numpy.ndarray[Block]
        :param max_world_version: The key to use to find the encoder.
        :type max_world_version: Tuple[str, Union[int, Tuple[int, int, int]]]
        :return: Raw data to be stored by the Format.
        :rtype: Any
        """
        raise NotImplementedError()

    def _encode_entity(
        self, entity: Entity, id_type: EntityIDType, coord_type: EntityCoordType
    ) -> Optional[amulet_nbt.NBTFile]:
        return self._encode_base_entity(entity, id_type, coord_type)

    def _encode_block_entity(
        self, entity: BlockEntity, id_type: EntityIDType, coord_type: EntityCoordType
    ) -> Optional[amulet_nbt.NBTFile]:
        return self._encode_base_entity(entity, id_type, coord_type)

    @staticmethod
    def _encode_base_entity(
        entity: Union[Entity, BlockEntity],
        id_type: EntityIDType,
        coord_type: EntityCoordType,
    ) -> Optional[amulet_nbt.NBTFile]:
        if not isinstance(entity.nbt, amulet_nbt.NBTFile) and isinstance(
            entity.nbt.value, amulet_nbt.TAG_Compound
        ):
            return
        nbt = entity.nbt

        if id_type == EntityIDType.namespace_str_id:
            nbt["id"] = amulet_nbt.TAG_String(entity.namespaced_name)
        elif id_type == EntityIDType.namespace_str_Id:
            nbt["Id"] = amulet_nbt.TAG_String(entity.namespaced_name)
        elif id_type == EntityIDType.namespace_str_identifier:
            nbt["identifier"] = amulet_nbt.TAG_String(entity.namespaced_name)
        elif id_type == EntityIDType.str_id:
            nbt["id"] = amulet_nbt.TAG_String(entity.base_name)
        elif id_type == EntityIDType.int_id:
            if not entity.base_name.isnumeric():
                return
            nbt["id"] = amulet_nbt.TAG_Int(int(entity.base_name))
        else:
            raise NotImplementedError(f"Entity id type {id_type}")

        if coord_type == EntityCoordType.Pos_list_double:
            nbt["Pos"] = amulet_nbt.TAG_List(
                [
                    amulet_nbt.TAG_Double(float(entity.x)),
                    amulet_nbt.TAG_Double(float(entity.y)),
                    amulet_nbt.TAG_Double(float(entity.z)),
                ]
            )
        elif coord_type == EntityCoordType.Pos_list_float:
            nbt["Pos"] = amulet_nbt.TAG_List(
                [
                    amulet_nbt.TAG_Float(float(entity.x)),
                    amulet_nbt.TAG_Float(float(entity.y)),
                    amulet_nbt.TAG_Float(float(entity.z)),
                ]
            )
        elif coord_type == EntityCoordType.Pos_list_int:
            nbt["Pos"] = amulet_nbt.TAG_List(
                [
                    amulet_nbt.TAG_Int(int(entity.x)),
                    amulet_nbt.TAG_Int(int(entity.y)),
                    amulet_nbt.TAG_Int(int(entity.z)),
                ]
            )
        elif coord_type == EntityCoordType.Pos_array_int:
            nbt["Pos"] = amulet_nbt.TAG_Int_Array(
                [int(entity.x), int(entity.y), int(entity.z)]
            )
        elif coord_type == EntityCoordType.xyz_int:
            nbt["x"] = amulet_nbt.TAG_Int(int(entity.x))
            nbt["y"] = amulet_nbt.TAG_Int(int(entity.y))
            nbt["z"] = amulet_nbt.TAG_Int(int(entity.z))
        else:
            raise NotImplementedError(f"Entity coord type {coord_type}")

        return nbt

    @staticmethod
    @overload
    def check_type(obj: TAG_Compound, key: str, dtype: Type[AnyNBT]) -> Any:
        ...

    @staticmethod
    @overload
    def check_type(obj: TAG_List, key: int, dtype: Type[AnyNBT]) -> Any:
        ...

    @staticmethod
    def check_type(obj: TAG_Compound, key: str, dtype: Type[AnyNBT]) -> bool:
        """Check the key exists and the type is correct."""
        return key in obj and type(obj[key]) is dtype

    @overload
    def get_obj(
        self,
        obj: TAG_Compound,
        key: str,
        dtype: Type[AnyNBT],
        default: Optional[AnyNBT] = None,
    ) -> Optional[AnyNBT]:
        ...

    @overload
    def get_obj(
        self,
        obj: TAG_List,
        key: int,
        dtype: Type[AnyNBT],
        default: Optional[AnyNBT] = None,
    ) -> Optional[AnyNBT]:
        ...

    def get_obj(
        self, obj, key, dtype: Type[AnyNBT], default: Optional[AnyNBT] = None
    ) -> Optional[AnyNBT]:
        """Pop a key from a container object if it exists and the type is correct. Otherwise return default.
        This works in much the same way as dict.get but uses default if the data type does not match.

        :param obj: The TAG_Compound to read from.
        :param key: The key to use.
        :param dtype: The expected data type.
        :param default: The default value to use if the existing is not valid. If None will use dtype()
        :return: The final value in the key.
        """
        if key in obj:
            if type(obj[key]) is dtype:
                # if it exists and is correct
                return obj.pop(key)
        if default is None:
            return dtype()
        return default

    @staticmethod
    def set_obj(
        obj: TAG_Compound,
        key: str,
        dtype: Type[AnyNBT],
        default: Optional[AnyNBT] = None,
    ) -> AnyNBT:
        """Set a key in a compound tag if the key does not exist or is not the correct type.
        This works in much the same way as dict.setdefault but overwrites if the data type does not match.

        :param obj: The TAG_Compound to apply to.
        :param key: The key to use.
        :param dtype: The expected data type.
        :param default: The default value to use if the existing is not valid. If None will use dtype()
        :return: The final value in the key.
        """
        if key not in obj or type(obj[key]) is not dtype:
            if default is None:
                obj[key] = dtype()
            else:
                assert type(default) is dtype
                obj[key] = default
        return obj[key]

    def get_translator(
        self,
        max_world_version: Tuple[str, Union[int, Tuple[int, int, int]]],
        data: Any = None,
    ) -> Tuple["Translator", Union[int, Tuple[int, int, int]]]:
        """
        Get the Translator class for the requested version.
        :param max_world_version: The game version the world was last opened in.
        :type max_world_version: Java: int (DataVersion)    Bedrock: Tuple[int, int, int, ...] (game version number)
        :param data: Optional data to get translator based on chunk version rather than world version
        :param data: Any
        :return: Tuple[Translator, version number for PyMCTranslate to use]
        :rtype: Tuple[Translator, Union[int, Tuple[int, int, int]]]
        """
        raise NotImplementedError

    @staticmethod
    def is_valid(key: Tuple) -> bool:
        """
        Returns whether this Interface is able to interface with the chunk type with a given identifier key,
        generated by the format.

        :param key: The key who's decodability needs to be checked.
        :return: True if this interface can interface with the chunk version associated with the key, False otherwise.
        """
        raise NotImplementedError()



from __future__ import annotations

from .leveldb_5 import (
    LevelDB5Interface,
)


class LevelDB6Interface(LevelDB5Interface):
    def __init__(self):
        super().__init__()

        self.features["chunk_version"] = 6


export = LevelDB6Interface



from typing import Optional, Any
from .changeable import Changeable

EntryKeyType = Any  # The key an entry is stored under
EntryType = Optional[
    Changeable
]  # None is reserved for if the entry was deleted/did not exist



from setuptools import setup, find_packages
import os
import glob
import shutil
import versioneer

# there were issues with other builds carrying over their cache
for d in glob.glob("*.egg-info"):
    shutil.rmtree(d)


def remove_git_and_http_package_links(uris):
    for uri in uris:
        if uri.startswith("git+") or uri.startswith("https:"):
            continue
        yield uri


with open("./requirements.txt") as requirements_fp:
    required_packages = [
        line for line in remove_git_and_http_package_links(requirements_fp.readlines())
    ]

package_data = [
    os.path.relpath(path, "amulet") for path in
    set(
        glob.glob(
            os.path.join(
                "amulet",
                "**",
                "*.*"
            ),
            recursive=True
        )
    ) - set(
        glob.glob(
            os.path.join(
                "amulet",
                "**",
                "*.py[cod]"
            ),
            recursive=True
        )
    )
]

setup(
    name="amulet-core",
    version=versioneer.get_version(),
    description="A Python library for reading/writing Minecraft's various save formats.",
    author="James Clare, Ben Gothard et al.",
    author_email="amuleteditor@gmail.com",
    install_requires=required_packages,
    packages=find_packages(),
    package_data={"amulet": package_data},
    cmdclass=versioneer.get_cmdclass(),
    setup_requires=required_packages,
    dependency_links=[
        "https://github.com/Amulet-Team/Amulet-NBT",
        "https://github.com/gentlegiantJGC/PyMCTranslate",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



from __future__ import annotations

from .leveldb_0 import (
    LevelDB0Interface,
)


class LevelDB1Interface(LevelDB0Interface):
    def __init__(self):
        super().__init__()

        self.features["chunk_version"] = 1


export = LevelDB1Interface



from __future__ import annotations

from .leveldb_10 import (
    LevelDB10Interface,
)


class LevelDB11Interface(LevelDB10Interface):
    def __init__(self):
        super().__init__()

        self.features["chunk_version"] = 11


export = LevelDB11Interface



from amulet.api.selection import SelectionGroup
from amulet.api.data_types import Dimension
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from amulet.api.level import BaseLevel


def delete_chunk(world: "BaseLevel", dimension: Dimension, source_box: SelectionGroup):
    iter_count = len(list(world.get_chunk_boxes(dimension, source_box)))
    count = 0

    for chunk, _ in world.get_chunk_boxes(dimension, source_box):
        cx, cz = chunk.coordinates
        world.delete_chunk(cx, cz, dimension)
        count += 1
        yield count / iter_count



from amulet.api.selection import SelectionGroup
from amulet.api.data_types import Dimension
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from amulet.api.level import BaseLevel


def clone(
    world: "BaseLevel", dimension: Dimension, selection: SelectionGroup, target: dict
):
    offset_x, offset_y, offset_z = (selection.max - selection.min) // 2
    dst_location = (
        target.get("x", 0) + offset_x,
        target.get("y", 0) + offset_y,
        target.get("z", 0) + offset_y,
    )
    world.paste(world, dimension, selection, dimension, dst_location)



from __future__ import annotations

from .leveldb_19 import (
    LevelDB19Interface,
)


class LevelDB20Interface(LevelDB19Interface):
    def __init__(self):
        super().__init__()

        self.features["chunk_version"] = 20


export = LevelDB20Interface



from __future__ import annotations

import os
import struct
from typing import Tuple, Any, Dict, Generator, Optional, List, TYPE_CHECKING, Union
import time
import glob
import shutil

import amulet_nbt as nbt

from amulet.api.wrapper import WorldFormatWrapper, DefaultVersion
from amulet.utils.format_utils import check_all_exist, load_leveldat
from amulet.api.errors import LevelDoesNotExist, ObjectWriteError, ChunkLoadError
from amulet.api.data_types import (
    ChunkCoordinates,
    VersionNumberInt,
    PlatformType,
    DimensionCoordinates,
)
from .dimension import AnvilDimensionManager

if TYPE_CHECKING:
    from amulet.api.data_types import Dimension
    from amulet.api.level import BaseLevel

InternalDimension = str


class AnvilFormat(WorldFormatWrapper):
    def __init__(self, directory: str):
        super().__init__(directory)
        self._platform = "java"
        self._root_tag: nbt.NBTFile = nbt.NBTFile()
        self._levels: Dict[InternalDimension, AnvilDimensionManager] = {}
        self._dimension_name_map: Dict["Dimension", InternalDimension] = {}
        self._mcc_support: Optional[bool] = None
        self._lock_time = None
        self._shallow_load()

    def _shallow_load(self):
        try:
            self._load_level_dat()
        except:
            pass

    def _load_level_dat(self):
        """Load the level.dat file and check the image file"""
        if os.path.isfile(os.path.join(self.path, "icon.png")):
            self._world_image_path = os.path.join(self.path, "icon.png")
        else:
            self._world_image_path = self._missing_world_icon
        self.root_tag = nbt.load(os.path.join(self.path, "level.dat"))

    @staticmethod
    def is_valid(directory) -> bool:
        """
        Returns whether this format is able to load a given world.

        :param directory: The path to the root of the world to load.
        :return: True if the world can be loaded by this format, False otherwise.
        """
        if not check_all_exist(directory, "level.dat"):
            return False

        try:
            level_dat_root = load_leveldat(directory)
        except:
            return False

        if "Data" not in level_dat_root:
            return False

        if "FML" in level_dat_root:
            return False

        return True

    @property
    def valid_formats(self) -> Dict[PlatformType, Tuple[bool, bool]]:
        return {"java": (True, True)}

    @property
    def version(self) -> VersionNumberInt:
        """The version number for the given platform the data is stored in eg (1, 16, 2)"""
        if self._version == DefaultVersion:
            self._version = self._get_version()
        return self._version

    def _get_version(self) -> VersionNumberInt:
        return (
            self.root_tag.get("Data", nbt.TAG_Compound())
            .get("DataVersion", nbt.TAG_Int(-1))
            .value
        )

    @property
    def root_tag(self) -> nbt.NBTFile:
        return self._root_tag

    @root_tag.setter
    def root_tag(self, root_tag: Union[nbt.NBTFile, nbt.TAG_Compound]):
        if type(root_tag) is nbt.TAG_Compound:
            self._root_tag = nbt.NBTFile(root_tag)
        elif type(root_tag) is nbt.NBTFile:
            self._root_tag = root_tag
        else:
            raise ValueError("root_tag must be a TAG_Compound or NBTFile")

    @property
    def world_name(self) -> str:
        """The name of the world"""
        return str(self.root_tag["Data"].get("LevelName", ""))

    @world_name.setter
    def world_name(self, value: str):
        self.root_tag["Data"]["LevelName"] = nbt.TAG_String(value)

    @property
    def last_played(self) -> int:
        return self.root_tag["Data"]["LastPlayed"].value

    @property
    def game_version_string(self) -> str:
        try:
            return f'Java {self.root_tag["Data"]["Version"]["Name"].value}'
        except Exception:
            return f"Java Unknown Version"

    @property
    def dimensions(self) -> List["Dimension"]:
        """A list of all the levels contained in the world"""
        return list(self._dimension_name_map.keys())

    def register_dimension(
        self,
        relative_dimension_path: InternalDimension,
        dimension_name: Optional["Dimension"] = None,
    ):
        """
        Register a new dimension.
        :param relative_dimension_path: The relative path to the dimension directory from the world root. "" for the world root.
        :param dimension_name: The name of the dimension shown to the user
        :return:
        """
        if dimension_name is None:
            dimension_name: "Dimension" = relative_dimension_path

        if relative_dimension_path:
            path = os.path.join(self.path, relative_dimension_path)
        else:
            path = self.path

        if (
            relative_dimension_path not in self._levels
            and dimension_name not in self._dimension_name_map
        ):
            self._levels[relative_dimension_path] = AnvilDimensionManager(
                path, mcc=self._mcc_support
            )
            self._dimension_name_map[dimension_name] = relative_dimension_path

    def _get_interface_key(
        self, raw_chunk_data: Optional[Any] = None
    ) -> Tuple[str, int]:
        if raw_chunk_data:
            return (
                self.platform,
                raw_chunk_data.get("DataVersion", nbt.TAG_Int(-1)).value,
            )
        else:
            return self.max_world_version

    def _reload_world(self):
        # reload the level.dat in case it has changed
        self._load_level_dat()

        # create the session.lock file (this has mostly been lifted from MCEdit)
        self._lock_time = int(time.time() * 1000)
        try:
            with open(os.path.join(self.path, "session.lock"), "wb") as f:
                f.write(struct.pack(">Q", self._lock_time))
                f.flush()
                os.fsync(f.fileno())
        except PermissionError as e:
            raise PermissionError(
                f"Could not access session.lock. The world may be open somewhere else.\n{e}"
            )

        self._mcc_support = (
            self.version > 2203
        )  # the real number might actually be lower

        # load all the levels
        self.register_dimension("", "overworld")
        self.register_dimension("DIM-1", "nether")
        self.register_dimension("DIM1", "end")

        for dir_name in os.listdir(self.path):
            level_path = os.path.join(self.path, dir_name)
            if os.path.isdir(level_path) and dir_name.startswith("DIM"):
                if AnvilDimensionManager.level_regex.fullmatch(dir_name) is None:
                    continue
                self.register_dimension(dir_name)

        for dimension_path in glob.glob(
            os.path.join(self.path, "dimensions", "*", "*", "region")
        ):
            dimension_path_split = dimension_path.split(os.sep)
            dimension_name = f"{dimension_path_split[-3]}/{dimension_path_split[-2]}"
            self.register_dimension(
                os.path.dirname(os.path.relpath(dimension_path, self.path)),
                dimension_name,
            )

    def _open(self):
        """Open the database for reading and writing"""
        self._reload_world()

    def _create(self, overwrite: bool, **kwargs):
        if os.path.isdir(self.path):
            if overwrite:
                shutil.rmtree(self.path)
            else:
                raise ObjectWriteError(
                    f"A world already exists at the path {self.path}"
                )
        self._version = self.translation_manager.get_version(
            self.platform, self.version
        ).data_version

        self.root_tag = root = nbt.TAG_Compound()
        root["Data"] = data = nbt.TAG_Compound()
        data["version"] = nbt.TAG_Int(19133)
        data["DataVersion"] = nbt.TAG_Int(self._version)
        data["LastPlayed"] = nbt.TAG_Long(int(time.time() * 1000))
        data["LevelName"] = nbt.TAG_String("World Created By Amulet")

        os.makedirs(self.path, exist_ok=True)
        self.root_tag.save_to(os.path.join(self.path, "level.dat"))
        self._reload_world()

    @property
    def has_lock(self) -> bool:
        """Verify that the world database can be read and written"""
        if self._has_lock:
            if self._lock_time is None:
                # the world was created not opened
                return True
            try:
                with open(os.path.join(self.path, "session.lock"), "rb") as f:
                    return struct.unpack(">Q", f.read(8))[0] == self._lock_time
            except:
                return False
        return False

    def pre_save_operation(self, level: "BaseLevel") -> Generator[float, None, bool]:
        """Logic to run before saving. Eg recalculating height maps or lighting.
        Must be a generator that yields a number and returns a bool.
        The yielded number is the progress from 0 to 1.
        The returned bool is if changes have been made.
        :param level: The level to apply modifications to.
        :return: Have any modifications been made.
        """
        changed_chunks = list(level.chunks.changed_chunks())
        height = self._calculate_height(level, changed_chunks)
        try:
            while True:
                yield next(height) / 2
        except StopIteration as e:
            height_changed = e.value

        light = self._calculate_light(level, changed_chunks)
        try:
            while True:
                yield next(light) / 2
        except StopIteration as e:
            light_changed = e.value

        return height_changed or light_changed

    @staticmethod
    def _calculate_height(
        level: "BaseLevel", chunks: List[DimensionCoordinates]
    ) -> Generator[float, None, bool]:
        """Calculate the height values for chunks."""
        chunk_count = len(chunks)
        # it looks like the game recalculates the height value if not defined.
        # Just delete the stored height values so that they do not get written back.
        # tested as of 1.12.2. This may not be true for older versions.
        changed = False
        for i, (dimension, cx, cz) in enumerate(chunks):
            try:
                chunk = level.get_chunk(cx, cz, dimension)
            except ChunkLoadError:
                pass
            else:
                changed_ = False
                changed_ |= chunk.misc.pop("height_mapC", None) is not None
                changed_ |= chunk.misc.pop("height_map256IA", None) is not None
                if changed_:
                    changed = True
                    chunk.changed = True
            yield i / chunk_count
        return changed

    @staticmethod
    def _calculate_light(
        level: "BaseLevel", chunks: List[DimensionCoordinates]
    ) -> Generator[float, None, bool]:
        """Calculate the height values for chunks."""
        # this is needed for before 1.14
        chunk_count = len(chunks)
        changed = False
        if level.level_wrapper.version < 1934:
            # the version may be less than 1934 but is at least 1924
            # calculate the light values
            pass
            # TODO
        else:
            # the game will recalculate the light levels
            for i, (dimension, cx, cz) in enumerate(chunks):
                try:
                    chunk = level.get_chunk(cx, cz, dimension)
                except ChunkLoadError:
                    pass
                else:
                    changed_ = False
                    changed_ |= chunk.misc.pop("block_light", None) is not None
                    changed_ |= chunk.misc.pop("sky_light", None) is not None
                    if changed_:
                        changed = True
                        chunk.changed = True
                yield i / chunk_count
        return changed

    def _save(self):
        """Save the data back to the disk database"""
        os.makedirs(self.path, exist_ok=True)
        for level in self._levels.values():
            level.save()
        self.root_tag.save_to(os.path.join(self.path, "level.dat"))
        # TODO: save other world data

    def _close(self):
        """Close the disk database"""
        pass

    def unload(self):
        """Unload data stored in the Format class"""
        for level in self._levels.values():
            level.unload()

    def _has_dimension(self, dimension: "Dimension"):
        return (
            dimension in self._dimension_name_map
            and self._dimension_name_map[dimension] in self._levels
        )

    def _get_dimension(self, dimension: "Dimension"):
        self._verify_has_lock()
        if self._has_dimension(dimension):
            return self._levels[self._dimension_name_map[dimension]]
        else:
            raise LevelDoesNotExist(dimension)

    def all_chunk_coords(
        self, dimension: "Dimension"
    ) -> Generator[ChunkCoordinates, None, None]:
        """A generator of all chunk coords in the given dimension"""
        if self._has_dimension(dimension):
            yield from self._get_dimension(dimension).all_chunk_coords()

    def has_chunk(self, cx: int, cz: int, dimension: Dimension) -> bool:
        return self._has_dimension(dimension) and self._get_dimension(
            dimension
        ).has_chunk(cx, cz)

    def _delete_chunk(self, cx: int, cz: int, dimension: "Dimension"):
        """Delete a chunk from a given dimension"""
        if self._has_dimension(dimension):
            self._get_dimension(dimension).delete_chunk(cx, cz)

    def _put_raw_chunk_data(self, cx: int, cz: int, data: Any, dimension: "Dimension"):
        """
        Actually stores the data from the interface to disk.
        """
        self._get_dimension(dimension).put_chunk_data(cx, cz, data)

    def _get_raw_chunk_data(
        self, cx: int, cz: int, dimension: "Dimension"
    ) -> nbt.NBTFile:
        """
        Return the raw data as loaded from disk.

        :param cx: The x coordinate of the chunk.
        :param cz: The z coordinate of the chunk.
        :param dimension: The dimension to load the data from.
        :return: The raw chunk data.
        """
        return self._get_dimension(dimension).get_chunk_data(cx, cz)


if __name__ == "__main__":
    import sys

    world_path = sys.argv[1]
    world = AnvilDimensionManager(world_path)
    chunk = world.get_chunk_data(0, 0)
    print(chunk)
    world.put_chunk_data(0, 0, chunk)
    world.save()



import numpy
from typing import List

import amulet_nbt

from amulet.api.selection import SelectionBox
from .data_types import BlockDataArrayType, BlockArrayType


class SchematicChunk:
    def __init__(
        self,
        selection: SelectionBox,
        blocks: BlockArrayType,
        data: BlockDataArrayType,
        block_entities: List[amulet_nbt.TAG_Compound],
        entities: List[amulet_nbt.TAG_Compound],
    ):
        self.selection = selection
        assert isinstance(blocks, numpy.ndarray) and blocks.dtype == numpy.uint16
        assert isinstance(data, numpy.ndarray) and data.dtype == numpy.uint8
        self.blocks = blocks
        self.data = data
        self.block_entities = block_entities
        self.entities = entities



from __future__ import annotations

from typing import Tuple, Union, TYPE_CHECKING
import numpy

from amulet.api.chunk import Chunk
from amulet.api.wrapper import Translator
from amulet.api.data_types import VersionIdentifierType, AnyNDArray, BlockNDArray
from amulet.api.registry import BlockManager

if TYPE_CHECKING:
    from PyMCTranslate import Version, TranslationManager


class JavaNumericalTranslator(Translator):
    def _translator_key(
        self, version_number: int
    ) -> Tuple[str, Union[int, Tuple[int, int, int]]]:
        return "java", version_number

    @staticmethod
    def _unpack_blocks(
        translation_manager: "TranslationManager",
        version_identifier: VersionIdentifierType,
        chunk: Chunk,
        block_palette: AnyNDArray,
    ):
        """
        Unpacks an int array of block ids and block data values [[1, 0], [2, 0]] into a numpy array of Block objects.
        :param translation_manager:
        :param version_identifier:
        :param chunk:
        :param block_palette:
        :return:
        """
        version = translation_manager.get_version(*version_identifier)
        chunk._block_palette = BlockManager(
            [version.block.ints_to_block(*entry) for entry in block_palette]
        )

    def _pack_block_palette(
        self, version: "Version", palette: BlockNDArray
    ) -> AnyNDArray:
        """
        Packs a numpy array of Block objects into an int array of block ids and block data values [[1, 0], [2, 0]].
        :param version:
        :param palette:
        :return:
        """
        palette = [version.block.block_to_ints(entry) for entry in palette]
        for index, value in enumerate(palette):
            if value is None:
                palette[index] = (
                    0,
                    0,
                )  # TODO: find some way for the user to specify this
        return numpy.array(palette)

    @staticmethod
    def is_valid(key):
        return key[0] == "java" and key[1] <= 1343


export = JavaNumericalTranslator



from collections import UserList
from typing import Iterable, Generator
import copy
from amulet.api.entity import Entity


class ChunkList(UserList):
    def __init__(self, iterable: Iterable = ()):
        super(ChunkList, self).__init__(list(iterable))

    def _check_all_types(self, value):
        [self._check_type(val) for val in value]

    def _check_type(self, value):
        raise NotImplementedError

    def append(self, value) -> None:
        """ Append value to the end of the list. """
        self._check_type(value)
        super().append(value)

    def copy(self) -> "ChunkList":
        return copy.deepcopy(self)

    def extend(self, iterable: Iterable[Entity]) -> None:
        """ Extend list by appending elements from the iterable. """
        self._check_all_types(iterable)
        super().extend(iterable)

    def insert(self, index: int, value) -> None:
        """ Insert value before index. """
        self._check_type(value)
        super().insert(index, value)

    def __repr__(self):
        if not self:
            return "[]"
        return "".join(f"\n\t{str(obj)}" for obj in self) + "\n"


class EntityList(ChunkList):
    def _check_type(self, value):
        assert isinstance(value, Entity)

    def __iter__(self) -> Generator[Entity, None, None]:
        yield from super().__iter__()

    def __repr__(self) -> str:
        """ Return repr(self). """
        return f"EntityList({super().__repr__()})"


if __name__ == "__main__":
    import amulet_nbt

    entities = EntityList()
    block_ents = [
        Entity("minecraft", "creeper", 0.0, 0.0, 0.0, amulet_nbt.NBTFile()),
        Entity("minecraft", "cow", 0.0, 0.0, 0.0, amulet_nbt.NBTFile()),
        Entity("minecraft", "pig", 0.0, 0.0, 0.0, amulet_nbt.NBTFile()),
        Entity("minecraft", "sheep", 0.0, 0.0, 0.0, amulet_nbt.NBTFile()),
    ]
    entities.append(Entity("minecraft", "cow", 0.0, 0.0, 0.0, amulet_nbt.NBTFile()))
    entities += block_ents
    print(entities)



import amulet_nbt


class Item(amulet_nbt.TAG_Compound):
    """
    {
        "namespace": TAG_String,
        "base_name": TAG_String,
        "metadata": TAG_Compound
    }
    """

    def __init__(self, namespace: str, base_name: str, metadata: dict = None):
        super().__init__(
            {
                "namespace": amulet_nbt.TAG_String(namespace),
                "base_name": amulet_nbt.TAG_String(base_name),
                "metadata": amulet_nbt.TAG_Compound(metadata or {}),
            }
        )

    @property
    def namespace(self) -> str:
        return self.value["namespace"].value

    @property
    def base_name(self) -> str:
        return self.value["base_name"].value

    @property
    def metadata(self) -> amulet_nbt.TAG_Compound:
        return self.value["metadata"].value


class BlockItem(amulet_nbt.TAG_Compound):
    """
    {
        "namespace": TAG_String,
        "base_name": TAG_String,
        "properties": TAG_Compound
        "metadata": TAG_Compound
    }
    """

    def __init__(
        self,
        namespace: str,
        base_name: str,
        properties: dict = None,
        metadata: dict = None,
    ):
        super().__init__(
            {
                "namespace": amulet_nbt.TAG_String(namespace),
                "base_name": amulet_nbt.TAG_String(base_name),
                "properties": amulet_nbt.TAG_Compound(properties or {}),
                "metadata": amulet_nbt.TAG_Compound(metadata or {}),
            }
        )

    @property
    def namespace(self) -> str:
        return self.value["namespace"].value

    @property
    def base_name(self) -> str:
        return self.value["base_name"].value

    @property
    def properties(self) -> amulet_nbt.TAG_Compound:
        return self.value["properties"].value

    @property
    def metadata(self) -> amulet_nbt.TAG_Compound:
        return self.value["metadata"].value



from __future__ import annotations

from .leveldb_3 import (
    LevelDB3Interface,
)


class LevelDB4Interface(LevelDB3Interface):
    def __init__(self):
        LevelDB3Interface.__init__(self)

        self.features["chunk_version"] = 4


export = LevelDB4Interface



# A cache for objects implemented using leveldb for speed.
import os
import shutil
import atexit

from amulet.libs.leveldb import LevelDB
from amulet.api.paths import get_temp_dir
from amulet import log

_path = os.path.join(get_temp_dir("cache"), "db")


_cache_db = None


def _clear_db():
    if _cache_db is not None:
        log.info("Removing cache.")
        _cache_db.close(compact=False)
        shutil.rmtree(_path)


def get_cache_db() -> LevelDB:
    global _cache_db
    if _cache_db is None:
        _cache_db = LevelDB(_path, True)
        atexit.register(_clear_db)
    return _cache_db



from __future__ import annotations

from typing import TYPE_CHECKING

from .base_level import BaseLevel

if TYPE_CHECKING:
    from amulet.api.wrapper import StructureFormatWrapper


class Structure(BaseLevel):
    """
    Class that handles editing of any structure format via an separate and flexible data format
    """

    def __init__(
        self,
        directory: str,
        structure_wrapper: "StructureFormatWrapper",
        temp_dir: str = None,
    ):
        super().__init__(directory, structure_wrapper, temp_dir)

    @property
    def level_wrapper(self) -> "StructureFormatWrapper":
        """A class to access data directly from the level."""
        return self._level_wrapper



from typing import BinaryIO, List, Any, Generator, Tuple, Dict, Optional, TYPE_CHECKING
import os

from .format_wrapper import FormatWrapper
from amulet.api.data_types import Dimension, PlatformType, ChunkCoordinates
from amulet.api.errors import ObjectReadError, ObjectReadWriteError

if TYPE_CHECKING:
    from amulet.api.wrapper import Interface


class StructureFormatWrapper(FormatWrapper):
    """A base FormatWrapper for all structures that only have one dimension."""

    def __init__(self, path: str):
        """Set up the StructureFormatWrapper ready for accessing.

        :param path: The location of the file to read from and write to.
        """
        if type(self) is StructureFormatWrapper:
            raise Exception(
                "StructureFormatWrapper is not directly usable. One of its subclasses must be used."
            )
        super().__init__(path)
        if self.extensions:
            ext = os.path.splitext(self._path)[1]
            if ext not in self.extensions:
                raise ObjectReadWriteError(
                    f"The given file does not have a valid file extension. Must be one of {self.extensions}"
                )

    @property
    def world_name(self) -> str:
        """The name of the world"""
        return os.path.basename(self.path)

    @property
    def dimensions(self) -> List[Dimension]:
        return ["main"]

    @property
    def can_add_dimension(self) -> bool:
        return False

    def register_dimension(self, dimension_internal: Any, dimension_name: Dimension):
        pass

    @property
    def requires_selection(self) -> bool:
        """Does this object require that a selection be defined when creating it from scratch?"""
        return True

    @staticmethod
    def is_valid(path: str) -> bool:
        raise NotImplementedError

    @property
    def valid_formats(self) -> Dict[PlatformType, Tuple[bool, bool]]:
        raise NotImplementedError

    @property
    def extensions(self) -> Tuple[str, ...]:
        """The extension the file can have to be valid.
        eg (".schematic",)"""
        raise NotImplementedError

    def _get_interface(self, raw_chunk_data: Optional[Any] = None) -> "Interface":
        raise NotImplementedError

    def _create(self, overwrite: bool, **kwargs):
        raise NotImplementedError

    def open_from(self, f: BinaryIO):
        """Load from a file like object. Useful to load from RAM rather than disk."""
        raise NotImplementedError

    def _open(self):
        if not os.path.isfile(self._path):
            raise ObjectReadError(f"There is no file to read at {self._path}")
        with open(self._path, "rb") as f:
            self.open_from(f)

    def save_to(self, f: BinaryIO):
        """Write to a file like object. Useful to write to RAM rather than disk.
        _save should open the file on disk and give it to this method."""
        raise NotImplementedError

    def _save(self):
        os.makedirs(os.path.dirname(self._path), exist_ok=True)
        with open(self._path, "wb") as f:
            self.save_to(f)

    def _close(self):
        raise NotImplementedError

    def unload(self):
        raise NotImplementedError

    def all_chunk_coords(
        self, dimension: Dimension
    ) -> Generator[ChunkCoordinates, None, None]:
        raise NotImplementedError

    def has_chunk(self, cx: int, cz: int, dimension: Dimension) -> bool:
        raise NotImplementedError

    def _delete_chunk(self, cx: int, cz: int, dimension: Dimension):
        raise NotImplementedError

    def _put_raw_chunk_data(self, cx: int, cz: int, data: Any, dimension: Dimension):
        raise NotImplementedError

    def _get_raw_chunk_data(self, cx: int, cz: int, dimension: Dimension) -> Any:
        raise NotImplementedError



from __future__ import annotations

from amulet.api.wrapper import EntityIDType
from .leveldb_8 import (
    LevelDB8Interface,
)


class LevelDB9Interface(LevelDB8Interface):
    def __init__(self):
        super().__init__()

        self.features["chunk_version"] = 9
        self.features[
            "entity_format"
        ] = (
            EntityIDType.namespace_str_identifier
        )  # EntityIDType.int_id is present until at least v7. Not sure which was present for v8


export = LevelDB9Interface



from .base_history import BaseHistory


class HistoryManager(BaseHistory):
    """The base class for all active history manager objects.
    The HistoryManager tracks which objects have changed but is not aware of how they have changed."""

    def undo(self):
        """Revert to the previous state."""
        raise NotImplementedError

    def redo(self):
        """Un-revert to the next state."""
        raise NotImplementedError

    def mark_saved(self):
        """Let the class know that the current state has been saved."""
        raise NotImplementedError

    @property
    def changed(self) -> bool:
        """Have there been modifications since the last save."""
        raise NotImplementedError

    def create_undo_point(self) -> bool:
        """
        Find what has changed since the last undo point and optionally create a new undo point.
        :return: Was an undo point created. If there were no changes no snapshot will be created.
        """
        raise NotImplementedError

    def restore_last_undo_point(self):
        """Restore the world to the state it was when self.create_undo_point was called.
        If an operation errors there may be modifications made that did not get tracked.
        This will revert those changes."""
        raise NotImplementedError

    def purge(self):
        """Unload all cached data. Effectively returns the class to its starting state."""
        raise NotImplementedError

    @property
    def undo_count(self) -> int:
        """The number of times the undo method can be run."""
        raise NotImplementedError

    @property
    def redo_count(self) -> int:
        """The number of times the redo method can be run."""
        raise NotImplementedError

    @property
    def unsaved_changes(self) -> int:
        """The number of changes that have been made since the last save"""
        raise NotImplementedError



from __future__ import annotations

from .leveldb_13 import (
    LevelDB13Interface,
)


class LevelDB14Interface(LevelDB13Interface):
    def __init__(self):
        super().__init__()

        self.features["chunk_version"] = 14


export = LevelDB14Interface



from __future__ import annotations

from .anvil_na import (
    AnvilNAInterface,
)


class Anvil112Interface(AnvilNAInterface):
    def __init__(self):
        AnvilNAInterface.__init__(self)

    @staticmethod
    def minor_is_valid(key: int):
        return 0 <= key < 1444


export = Anvil112Interface



import numpy
from typing import Union, Optional, Dict, Tuple
from copy import deepcopy


from amulet.api.partial_3d_array import UnboundedPartial3DArray


class Biomes3D(UnboundedPartial3DArray):
    def __init__(
        self,
        input_array: Optional[Union[Dict[int, numpy.ndarray], "Biomes3D"]] = None,
    ):
        if input_array is None:
            input_array = {}
        if isinstance(input_array, Biomes3D):
            input_array: dict = deepcopy(input_array._sections)
        if not isinstance(input_array, dict):
            raise Exception(f"Input array must be Biomes3D or dict, got {input_array}")
        super().__init__(numpy.uint32, 0, (4, 4, 4), (0, 16), sections=input_array)


class Biomes:
    # __slots__ = ("_2d", "_3d", "_dimension")

    def __init__(
        self, array: Union[numpy.ndarray, Biomes3D, Dict[int, numpy.ndarray]] = None
    ):
        self._2d: Optional[numpy.ndarray] = None
        self._3d: Optional[Biomes3D] = None
        if array is None:
            self._dimension = 0
        elif isinstance(array, numpy.ndarray):
            assert array.shape == (
                16,
                16,
            ), "If Biomes is given an ndarray it must be 16x16"
            self._2d = array.copy()
            self._dimension = 2
        elif isinstance(array, (dict, Biomes3D)):
            self._3d = Biomes3D(array)
            self._dimension = 3

    def to_raw(
        self,
    ) -> Tuple[int, Optional[numpy.ndarray], Optional[Dict[int, numpy.ndarray]]]:
        """Don't use this method. Use to pickle data."""
        if self._3d is None:
            sections = None
        else:
            sections = self._3d._sections
        return self._dimension, self._2d, sections

    @classmethod
    def from_raw(
        cls,
        dimension: int,
        d2: Optional[numpy.ndarray],
        d3: Optional[Dict[int, numpy.ndarray]],
    ) -> "Biomes":
        """Don't use this method. Use to unpickle data."""
        biomes = cls()
        biomes._dimension = dimension
        biomes._2d = d2
        if d3 is not None:
            biomes._3d = Biomes3D(d3)
        return biomes

    @property
    def dimension(self) -> int:
        """The number of dimensions the data has
        0 when there is no data. Will error if you try accessing data from it without converting to one of the other formats.
        2 when the data is a 2D 16x16 numpy array.
        3 when the data is a 3D 4xinfx4 Biome3D array made of sections of size 4x4x4."""
        return self._dimension

    def convert_to_2d(self):
        """Convert the data in whatever format it is in to the 2D 16x16 format."""
        if self._2d is None:
            self._2d = numpy.zeros((16, 16), numpy.uint32)
        if self._dimension != 2:
            if self._dimension == 3 and self._3d is not None:
                # convert from 3D
                self._2d[:, :] = numpy.kron(
                    numpy.reshape(self._3d[:, 0, :], (4, 4)), numpy.ones((4, 4))
                )
            self._dimension = 2

    def convert_to_3d(self):
        """Convert the data in whatever format it is in to the 3D 4xinfx4 format."""
        if self._3d is None:
            self._3d = Biomes3D()
        if self._dimension != 3:
            if self._dimension == 2 and self._2d is not None:
                # convert from 2D
                self._3d[:, 0, :] = self._2d[::4, ::4].reshape(4, 1, 4)
            self._dimension = 3

    def _get_active(self) -> Union[numpy.ndarray, Biomes3D]:
        if self._dimension == 0:
            raise Exception(
                "You are trying to use Biomes but have not picked a format. Use one of the convert methods to specify the format."
            )
        elif self._dimension == 2:
            self.convert_to_2d()
            return self._2d
        elif self._dimension == 3:
            self.convert_to_3d()
            return self._3d
        else:
            raise Exception("Dimension is invalid. This shouldn't happen")

    def copy(self):
        return self.__copy__()

    def __copy__(self):
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__.update(self.__dict__)
        return result

    def __deepcopy__(self, memodict=None):
        cls = self.__class__
        result = cls.__new__(cls)
        memodict[id(self)] = result
        for k, v in self.__dict__.items():
            setattr(result, k, deepcopy(v, memodict))
        return result

    def __getattr__(self, item):
        return self._get_active().__getattribute__(item)

    def __getitem__(self, item):
        return self._get_active().__getitem__(item)

    def __setitem__(self, key, value):
        self._get_active().__setitem__(key, value)
